@P lemma
@O lemma
@P tag
@O tag
@P form
@O form
@P afun
@O afun
@L1 afun|---|Pred|Pnom|AuxV|Sb|Obj|Atr|Adv|AtrAdv|AdvAtr|Coord|AtrObj|ObjAtr|AtrAtr|AuxT|AuxR|AuxP|Apos|ExD|AuxC|Atv|AtvV|AuxO|AuxZ|AuxY|AuxG|AuxK|AuxX|AuxS|Pred_Co|Pnom_Co|AuxV_Co|Sb_Co|Obj_Co|Atr_Co|Adv_Co|AtrAdv_Co|AdvAtr_Co|Coord_Co|AtrObj_Co|ObjAtr_Co|AtrAtr_Co|AuxT_Co|AuxR_Co|AuxP_Co|Apos_Co|ExD_Co|AuxC_Co|Atv_Co|AtvV_Co|AuxO_Co|AuxZ_Co|AuxY_Co|AuxG_Co|AuxK_Co|AuxX_Co|Pred_Ap|Pnom_Ap|AuxV_Ap|Sb_Ap|Obj_Ap|Atr_Ap|Adv_Ap|AtrAdv_Ap|AdvAtr_Ap|Coord_Ap|AtrObj_Ap|ObjAtr_Ap|AtrAtr_Ap|AuxT_Ap|AuxR_Ap|AuxP_Ap|Apos_Ap|ExD_Ap|AuxC_Ap|Atv_Ap|AtvV_Ap|AuxO_Ap|AuxZ_Ap|AuxY_Ap|AuxG_Ap|AuxK_Ap|AuxX_Ap|Pred_Pa|Pnom_Pa|AuxV_Pa|Sb_Pa|Obj_Pa|Atr_Pa|Adv_Pa|AtrAdv_Pa|AdvAtr_Pa|Coord_Pa|AtrObj_Pa|ObjAtr_Pa|AtrAtr_Pa|AuxT_Pa|AuxR_Pa|AuxP_Pa|Apos_Pa|ExD_Pa|AuxC_Pa|Atv_Pa|AtvV_Pa|AuxO_Pa|AuxZ_Pa|AuxY_Pa|AuxG_Pa|AuxK_Pa|AuxX_Pa|Generated|NA|???
@P ID1
@P ID2
@VA origf
@P origf
@P afunprev
@P semPOS
@P tagauto
@P lemauto
@N ord
@P dord
@W sentord
@P govTR
@P nospace
@P root
@P ending
@P punct
@P alltags
@P wt
@P origfkind
@P formtype
@P gappost
@P gappre
@P para
@P cstslang
@P cstssource
@P cstsmarkup
@P chap
@P doc
@P docid
@P docmarkup
@P docprolog
@P1 warning
@P3 err1
@P3 err2
@P reserve1
@P reserve2
@P reserve3
@P reserve4
@P reserve5
@P wMDl_b
@P wMDt_a
@P tagMD_a
@P wMDt_b
@P tagMD_b
@P lemmaMD_a
@P lemmaMD_b
@P wMDl_a

[#,Z#-------------,#1,AuxS,cmpr9410:001-p2s1,origf=#1,ord=0,sentord=0,para=2,cs,Českomoravský profit,<mauth>zu<mdate>1996-2000<mdesc>Man. anal. annot,<mauth>ej\,lk<mdate>1996-2000<mdesc>Man. morph. annot.,1,s/inf/j/1994/cmpr9410,001,docprolog=<mod>s<txtype>inf<genre>mix<med>j<temp>1994<authname>y<opus>cmpr9410<id>001]([unie,NNFS1-----A----,unie,ExD,origf=unie,ord=2,sentord=2,formtype=lower,tagMD_a=NNFS1-----A----,tagMD_b=NNFS1-----A----,unie,unie]([celní,AAFS1----1A----,Celní,Atr,origf=Celní,ord=1,sentord=1,formtype=cap,tagMD_a=AAFS1----1A----,tagMD_b=AAFS1----1A----,celní,celní],[v-1,RR--6----------,v,AuxP,origf=v,ord=3,sentord=3,formtype=lower,tagMD_a=RR--6----------,tagMD_b=RR--6----------,v-1,v-1]([ohrožení_^(*8hrozit),NNNS6-----A----,ohrožení,Atr,origf=ohrožení,ord=4,sentord=4,formtype=lower,tagMD_a=NNNS6-----A----,tagMD_b=NNNS6-----A----,ohrožení_^(*8hrozit),ohrožení_^(*8hrozit)])))
[#,Z#-------------,#2,AuxS,cmpr9410:001-p3s1,origf=#2,ord=0,sentord=0,gappost=<i>b,para=3]([omezení_^(*3it),NNNP1-----A----,omezení,ExD,origf=omezení,ord=3,sentord=3,formtype=lower,tagMD_a=NNNS1-----A----,tagMD_b=NNNS1-----A----,omezení_^(*3it),omezení_^(*3it)]([nový,AANP1----1A----,Nová,Atr,origf=Nová,ord=1,sentord=1,formtype=cap,tagMD_a=AAFS1----1A----,tagMD_b=AAFS1----1A----,nový,nový],[striktní,AANP1----1A----,striktní,Atr,origf=striktní,ord=2,sentord=2,formtype=lower,tagMD_a=AANS1----1A----,tagMD_b=AANS1----1A----,striktní,striktní],[vláda,NNFS2-----A----,vlády,Atr,origf=vlády,ord=4,sentord=4,formtype=lower,tagMD_a=NNFS2-----A----,tagMD_b=NNFS2-----A----,vláda,vláda]([SR-1_:B_;G_^(Slovenská_republika),NNFXX-----A---8,SR,Atr,origf=SR,ord=5,sentord=5,formtype=upper,tagMD_a=NNFXX-----A---8,tagMD_b=NNFXX-----A---8,SR-1_:B_;G_^(Slovenská_republika),SR-1_:B_;G_^(Slovenská_republika)]),[proti-1,RR--3----------,proti,AuxP,origf=proti,ord=6,sentord=6,formtype=lower,tagMD_a=RR--3----------,tagMD_b=RR--3----------,proti-1,proti-1]([exportér,NNMP3-----A----,exportérům,Atr,origf=exportérům,ord=8,sentord=8,formtype=lower,tagMD_a=NNMP3-----A----,tagMD_b=NNMP3-----A----,exportér,exportér]([český,AAMP3----1A----,českým,Atr,origf=česklým,ord=7,sentord=7,origfkind=spell,lower,tagMD_a=AAMP3----1A----,tagMD_b=AAMP3----1A----,český,český]))))
(2,3)
//Tred:Custom-Attribute:${form}
//Tred:Custom-Attribute:${afun}
//Tred:Balloon-Pattern:tag:	${tag}
