<?xml version="1.0" encoding="utf-8"?>
<tdata xmlns="http://ufal.mff.cuni.cz/pdt/pml/">
 <head>
  <schema href="tdata_schema.xml"/>
  <references>
   <reffile id="a" name="adata" href="sample0.a"/>
   <reffile id="v" name="vallex" href="vallex.xml"/>
  </references>
 </head>
 <meta>
  <annotation_info>
   <desc>Manual annotation</desc>
  </annotation_info>
 </meta>
 <trees>
  <LM id="t-ln94210-2-p1s1">
   <atree.rf>a#a-ln94210-2-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-2-p1s1w1">
     <a>
      <lex.rf>a#a-ln94210-2-p1s1w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>čtvrť</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>fem</gender>
      <number>sg</number>
     </gram>
     <children>
      <LM id="t-ln94210-2-p1s1w3">
       <a>
        <lex.rf>a#a-ln94210-2-p1s1w3</lex.rf>
        <aux.rf>a#a-ln94210-2-p1s1w2</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>diplomat</t_lemma>
       <functor>BEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>pl</number>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-2-p2s1">
   <atree.rf>a#a-ln94210-2-p2s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-2-p2s1w24">
     <a>
      <lex.rf>a#a-ln94210-2-p2s1w24</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>připravovat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>22</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w5241f2</val_frame.rf>
     <children>
      <LM id="t-ln94210-2-p2s1w1">
       <a>
        <lex.rf>a#a-ln94210-2-p2s1w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>výstavba</t_lemma>
       <functor>PAT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-2-p2s1w3">
         <a>
          <lex.rf>a#a-ln94210-2-p2s1w3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>čtvrť</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-2-p2s1w2">
           <a>
            <lex.rf>a#a-ln94210-2-p2s1w2</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>diplomatický</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>3</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
          <LM id="t-ln94210-2-p2s1w6">
           <a>
            <lex.rf>a#a-ln94210-2-p2s1w6</lex.rf>
            <aux.rf>a#a-ln94210-2-p2s1w4</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>město</t_lemma>
           <functor>LOC</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>4</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-2-p2s1w5">
             <a>
              <lex.rf>a#a-ln94210-2-p2s1w5</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>hlavní</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>5</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-2-p2s1w11">
           <a>
            <lex.rf>a#a-ln94210-2-p2s1w11</lex.rf>
            <aux.rf>a#a-ln94210-2-p2s1w7</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>dům</t_lemma>
           <functor>ACMP</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>pl</number>
           </gram>
           <children>
            <LM id="t-ln94210-2-p2s1w10">
             <a>
              <lex.rf>a#a-ln94210-2-p2s1w10</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>rodinný</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>7</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
            <LM id="t-ln94210-2-p2s1w9">
             <a>
              <lex.rf>a#a-ln94210-2-p2s1w9</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>400</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>8</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
             <children>
              <LM id="t-ln94210-2-p2s1w8">
               <a>
                <lex.rf>a#a-ln94210-2-p2s1w8</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>zhruba</t_lemma>
               <functor>EXT</functor>
               <subfunctor>basic</subfunctor>
               <tfa>f</tfa>
               <deepord>9</deepord>
               <gram>
                <sempos>adv.denot.ngrad.nneg</sempos>
               </gram>
              </LM>
             </children>
            </LM>
            <LM id="t-ln94210-2-p2s1w13">
             <a>
              <lex.rf>a#a-ln94210-2-p2s1w13</lex.rf>
              <aux.rf>a#a-ln94210-2-p2s1w12</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>ubytování</t_lemma>
             <functor>AIM</functor>
             <tfa>f</tfa>
             <deepord>12</deepord>
             <gram>
              <sempos>n.denot.neg</sempos>
              <gender>neut</gender>
              <number>sg</number>
              <negation>neg0</negation>
             </gram>
             <val_frame.rf>v#v-w6979f1</val_frame.rf>
             <children>
              <LM id="t-ln94210-2-p2s1a1">
               <nodetype>qcomplex</nodetype>
               <is_generated>1</is_generated>
               <t_lemma>#Gen</t_lemma>
               <functor>ACT</functor>
               <tfa>t</tfa>
               <deepord>10</deepord>
              </LM>
              <LM id="t-ln94210-2-p2s1a0">
               <nodetype>qcomplex</nodetype>
               <is_generated>1</is_generated>
               <t_lemma>#Oblfm</t_lemma>
               <functor>LOC</functor>
               <subfunctor>nr</subfunctor>
               <tfa>t</tfa>
               <deepord>11</deepord>
              </LM>
              <LM id="t-ln94210-2-p2s1w18">
               <a>
                <lex.rf>a#a-ln94210-2-p2s1w18</lex.rf>
               </a>
               <nodetype>coap</nodetype>
               <t_lemma>a</t_lemma>
               <functor>GRAD</functor>
               <deepord>16</deepord>
               <children>
                <LM id="t-ln94210-2-p2s1w14">
                 <a>
                  <lex.rf>a#a-ln94210-2-p2s1w14</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>pracovník</t_lemma>
                 <functor>PAT</functor>
                 <is_member>1</is_member>
                 <tfa>f</tfa>
                 <deepord>13</deepord>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>anim</gender>
                  <number>pl</number>
                 </gram>
                 <children>
                  <LM id="t-ln94210-2-p2s1w16">
                   <a>
                    <lex.rf>a#a-ln94210-2-p2s1w16</lex.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>úřad</t_lemma>
                   <functor>APP</functor>
                   <tfa>f</tfa>
                   <deepord>14</deepord>
                   <gram>
                    <sempos>n.denot</sempos>
                    <gender>inan</gender>
                    <number>pl</number>
                   </gram>
                   <children>
                    <LM id="t-ln94210-2-p2s1w15">
                     <a>
                      <lex.rf>a#a-ln94210-2-p2s1w15</lex.rf>
                     </a>
                     <nodetype>complex</nodetype>
                     <t_lemma>zastupitelský</t_lemma>
                     <functor>RSTR</functor>
                     <tfa>f</tfa>
                     <deepord>15</deepord>
                     <gram>
                      <sempos>adj.denot</sempos>
                      <degcmp>pos</degcmp>
                      <negation>neg0</negation>
                     </gram>
                    </LM>
                   </children>
                  </LM>
                 </children>
                </LM>
                <LM id="t-ln94210-2-p2s1w19">
                 <a>
                  <lex.rf>a#a-ln94210-2-p2s1w19</lex.rf>
                 </a>
                 <nodetype>atom</nodetype>
                 <t_lemma>zejména</t_lemma>
                 <functor>CM</functor>
                 <deepord>17</deepord>
                </LM>
                <LM id="t-ln94210-2-p2s1w20">
                 <a>
                  <lex.rf>a#a-ln94210-2-p2s1w20</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>zaměstnanec</t_lemma>
                 <functor>PAT</functor>
                 <is_member>1</is_member>
                 <tfa>f</tfa>
                 <deepord>18</deepord>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>anim</gender>
                  <number>pl</number>
                 </gram>
                 <children>
                  <LM id="t-ln94210-2-p2s1w21">
                   <a>
                    <lex.rf>a#a-ln94210-2-p2s1w21</lex.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>rádio</t_lemma>
                   <functor>APP</functor>
                   <tfa>f</tfa>
                   <deepord>19</deepord>
                   <gram>
                    <sempos>n.denot</sempos>
                    <gender>neut</gender>
                    <number>sg</number>
                   </gram>
                   <children>
                    <LM id="t-ln94210-2-p2s1w23">
                     <a>
                      <lex.rf>a#a-ln94210-2-p2s1w23</lex.rf>
                     </a>
                     <nodetype>complex</nodetype>
                     <t_lemma>Evropa</t_lemma>
                     <functor>ID</functor>
                     <tfa>f</tfa>
                     <deepord>20</deepord>
                     <gram>
                      <sempos>n.denot</sempos>
                      <gender>fem</gender>
                      <number>sg</number>
                     </gram>
                     <children>
                      <LM id="t-ln94210-2-p2s1w22">
                       <a>
                        <lex.rf>a#a-ln94210-2-p2s1w22</lex.rf>
                       </a>
                       <nodetype>complex</nodetype>
                       <t_lemma>Svobodný</t_lemma>
                       <functor>RSTR</functor>
                       <tfa>f</tfa>
                       <deepord>21</deepord>
                       <gram>
                        <sempos>adj.denot</sempos>
                        <degcmp>pos</degcmp>
                        <negation>neg0</negation>
                       </gram>
                      </LM>
                     </children>
                    </LM>
                   </children>
                  </LM>
                 </children>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-2-p2s1w26">
       <a>
        <lex.rf>a#a-ln94210-2-p2s1w26</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>radnice</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>23</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-2-p2s1w25">
         <a>
          <lex.rf>a#a-ln94210-2-p2s1w25</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>pražský</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>24</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-2-p3s1">
   <atree.rf>a#a-ln94210-2-p3s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-2-p3s1w11">
     <a>
      <lex.rf>a#a-ln94210-2-p3s1w11</lex.rf>
      <aux.rf>
       <LM>a#a-ln94210-2-p3s1w7</LM>
       <LM>a#a-ln94210-2-p3s1w10</LM>
      </aux.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>vzniknout</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>8</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>cdn</verbmod>
      <deontmod>hrt</deontmod>
      <tense>sim</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w8671f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-2-p3s1w2">
       <a>
        <lex.rf>a#a-ln94210-2-p3s1w2</lex.rf>
        <aux.rf>a#a-ln94210-2-p3s1w1</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>slovo</t_lemma>
       <functor>CRIT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln94210-2-p3s1w6">
         <a>
          <lex.rf>a#a-ln94210-2-p3s1w6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Koukal</t_lemma>
         <functor>ACT</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <is_name_of_person>1</is_name_of_person>
         <gram>
          <sempos>n.denot</sempos>
          <gender>anim</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-2-p3s1w5">
           <a>
            <lex.rf>a#a-ln94210-2-p3s1w5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Jan</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>3</deepord>
           <is_name_of_person>1</is_name_of_person>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
          </LM>
          <LM id="t-ln94210-2-p3s1w4">
           <a>
            <lex.rf>a#a-ln94210-2-p3s1w4</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>primátor</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>4</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-2-p3s1w3">
             <a>
              <lex.rf>a#a-ln94210-2-p3s1w3</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>pražský</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>5</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-2-p3s1w9">
       <a>
        <lex.rf>a#a-ln94210-2-p3s1w9</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>čtvrť</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>7</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-2-p3s1w8">
         <a>
          <lex.rf>a#a-ln94210-2-p3s1w8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>tento</t_lemma>
         <functor>RSTR</functor>
         <tfa>t</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>adj.pron.def.demon</sempos>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-2-p3s1w14">
       <a>
        <lex.rf>a#a-ln94210-2-p3s1w14</lex.rf>
       </a>
       <nodetype>coap</nodetype>
       <t_lemma>a</t_lemma>
       <functor>CONJ</functor>
       <deepord>10</deepord>
       <children>
        <LM id="t-ln94210-2-p3s1w13">
         <a>
          <lex.rf>a#a-ln94210-2-p3s1w13</lex.rf>
          <aux.rf>a#a-ln94210-2-p3s1w12</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>rok</t_lemma>
         <functor>TPAR</functor>
         <is_member>1</is_member>
         <tfa>f</tfa>
         <deepord>9</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-2-p3s1w15">
         <a>
          <lex.rf>a#a-ln94210-2-p3s1w15</lex.rf>
          <aux.rf>a#a-ln94210-2-p3s1w12</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>půl</t_lemma>
         <functor>TPAR</functor>
         <is_member>1</is_member>
         <tfa>f</tfa>
         <deepord>11</deepord>
         <gram>
          <sempos>n.quant.def</sempos>
          <gender>nr</gender>
          <number>pl</number>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-2-p3s2">
   <atree.rf>a#a-ln94210-2-p3s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-2-p3s2w2">
     <a>
      <lex.rf>a#a-ln94210-2-p3s2w2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>hledat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w1058f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-2-p3s2w1">
       <a>
        <lex.rf>a#a-ln94210-2-p3s2w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>radnice</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln94210-2-p3s2w4">
       <a>
        <lex.rf>a#a-ln94210-2-p3s2w4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>místo</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-2-p3s2w3">
         <a>
          <lex.rf>a#a-ln94210-2-p3s2w3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>vhodný</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
         <children>
          <LM id="t-ln94210-2-p3s2w8">
           <a>
            <lex.rf>a#a-ln94210-2-p3s2w8</lex.rf>
            <aux.rf>
             <LM>a#a-ln94210-2-p3s2w6</LM>
             <LM>a#a-ln94210-2-p3s2w7</LM>
             <LM>a#a-ln94210-2-p3s2w5</LM>
            </aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>bezpečnost</t_lemma>
           <functor>REG</functor>
           <tfa>f</tfa>
           <deepord>5</deepord>
           <gram>
            <sempos>n.denot.neg</sempos>
            <gender>fem</gender>
            <number>sg</number>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-2-p3s3">
   <atree.rf>a#a-ln94210-2-p3s3</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-2-p3s3w13">
     <a>
      <lex.rf>a#a-ln94210-2-p3s3w13</lex.rf>
     </a>
     <nodetype>coap</nodetype>
     <t_lemma>a</t_lemma>
     <functor>CONJ</functor>
     <deepord>15</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-ln94210-2-p3s3w3">
       <a>
        <lex.rf>a#a-ln94210-2-p3s3w3</lex.rf>
        <aux.rf>a#a-ln94210-2-p3s3w1</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>ujednání</t_lemma>
       <functor>CRIT</functor>
       <tfa>t</tfa>
       <deepord>4</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>neut</gender>
        <number>pl</number>
        <negation>neg0</negation>
       </gram>
       <val_frame.rf>v#v-w7074f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-2-p3s3a0">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>1</deepord>
        </LM>
        <LM id="t-ln94210-2-p3s3a2">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>PAT</functor>
         <tfa>t</tfa>
         <deepord>2</deepord>
        </LM>
        <LM id="t-ln94210-2-p3s3a1">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ADDR</functor>
         <tfa>t</tfa>
         <deepord>3</deepord>
        </LM>
        <LM id="t-ln94210-2-p3s3w2">
         <a>
          <lex.rf>a#a-ln94210-2-p3s3w2</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>předběžný</t_lemma>
         <functor>TWHEN</functor>
         <subfunctor>nr</subfunctor>
         <tfa>f</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-2-p3s3w7">
       <a>
        <lex.rf>a#a-ln94210-2-p3s3w7</lex.rf>
        <aux.rf>a#a-ln94210-2-p3s3w4</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>poskytnout</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>9</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>cdn</verbmod>
        <deontmod>decl</deontmod>
        <tense>sim</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w3973f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-2-p3s3w6">
         <a>
          <lex.rf>a#a-ln94210-2-p3s3w6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>město</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-2-p3s3w5">
           <a>
            <lex.rf>a#a-ln94210-2-p3s3w5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>hlavní</t_lemma>
           <functor>RSTR</functor>
           <tfa>t</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-2-p3s3a3">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ADDR</functor>
         <tfa>t</tfa>
         <deepord>8</deepord>
        </LM>
        <LM id="t-ln94210-2-p3s3w9">
         <a>
          <lex.rf>a#a-ln94210-2-p3s3w9</lex.rf>
          <aux.rf>a#a-ln94210-2-p3s3w8</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>vybudování</t_lemma>
         <functor>AIM</functor>
         <tfa>f</tfa>
         <deepord>12</deepord>
         <gram>
          <sempos>n.denot.neg</sempos>
          <gender>neut</gender>
          <number>sg</number>
          <negation>neg0</negation>
         </gram>
         <val_frame.rf>v#v-w7799f1</val_frame.rf>
         <children>
          <LM id="t-ln94210-2-p3s3w10">
           <a>
            <lex.rf>a#a-ln94210-2-p3s3w10</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>čtvrť</t_lemma>
           <functor>PAT</functor>
           <tfa>t</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
          </LM>
          <LM id="t-ln94210-2-p3s3a4">
           <nodetype>qcomplex</nodetype>
           <is_generated>1</is_generated>
           <t_lemma>#Gen</t_lemma>
           <functor>ACT</functor>
           <tfa>t</tfa>
           <deepord>11</deepord>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-2-p3s3w12">
         <a>
          <lex.rf>a#a-ln94210-2-p3s3w12</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>pozemek</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>14</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-2-p3s3w11">
           <a>
            <lex.rf>a#a-ln94210-2-p3s3w11</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>#PersPron</t_lemma>
           <functor>APP</functor>
           <tfa>t</tfa>
           <deepord>13</deepord>
           <gram>
            <sempos>n.pron.def.pers</sempos>
            <gender>inher</gender>
            <number>inher</number>
            <person>inher</person>
            <politeness>inher</politeness>
           </gram>
           <coref_gram.rf>t-ln94210-2-p3s3w6</coref_gram.rf>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-2-p3s3w17">
       <a>
        <lex.rf>a#a-ln94210-2-p3s3w17</lex.rf>
        <aux.rf>a#a-ln94210-2-p3s3w16</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>zajistit</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>18</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>cdn</verbmod>
        <deontmod>decl</deontmod>
        <tense>sim</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w8845f4</val_frame.rf>
       <children>
        <LM id="t-ln94210-2-p3s3w15">
         <a>
          <lex.rf>a#a-ln94210-2-p3s3w15</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>vláda</t_lemma>
         <functor>ACT</functor>
         <tfa>c</tfa>
         <deepord>16</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-2-p3s3w14">
           <a>
            <lex.rf>a#a-ln94210-2-p3s3w14</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>český</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>17</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-2-p3s3w18">
         <a>
          <lex.rf>a#a-ln94210-2-p3s3w18</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>půjčka</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>20</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-2-p3s3a5">
           <nodetype>qcomplex</nodetype>
           <is_generated>1</is_generated>
           <t_lemma>#Gen</t_lemma>
           <functor>PAT</functor>
           <tfa>t</tfa>
           <deepord>19</deepord>
          </LM>
          <LM id="t-ln94210-2-p3s3w20">
           <a>
            <lex.rf>a#a-ln94210-2-p3s3w20</lex.rf>
            <aux.rf>a#a-ln94210-2-p3s3w19</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>financování</t_lemma>
           <functor>AIM</functor>
           <tfa>f</tfa>
           <deepord>22</deepord>
           <gram>
            <sempos>n.denot.neg</sempos>
            <gender>neut</gender>
            <number>sg</number>
            <negation>neg0</negation>
           </gram>
           <val_frame.rf>v#v-w953f1</val_frame.rf>
           <children>
            <LM id="t-ln94210-2-p3s3a6">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#Gen</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>21</deepord>
            </LM>
            <LM id="t-ln94210-2-p3s3w22">
             <a>
              <lex.rf>a#a-ln94210-2-p3s3w22</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>práce</t_lemma>
             <functor>PAT</functor>
             <tfa>f</tfa>
             <deepord>23</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>pl</number>
             </gram>
             <children>
              <LM id="t-ln94210-2-p3s3w21">
               <a>
                <lex.rf>a#a-ln94210-2-p3s3w21</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>stavební</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>24</deepord>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>pos</degcmp>
                <negation>neg0</negation>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-2-p3s4">
   <atree.rf>a#a-ln94210-2-p3s4</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-2-p3s4w7">
     <a>
      <lex.rf>a#a-ln94210-2-p3s4w7</lex.rf>
      <aux.rf>a#a-ln94210-2-p3s4w3</aux.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>splatit</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>7</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>hrt</deontmod>
      <tense>sim</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w6342f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-2-p3s4w2">
       <a>
        <lex.rf>a#a-ln94210-2-p3s4w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>strana</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-2-p3s4w1">
         <a>
          <lex.rf>a#a-ln94210-2-p3s4w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>americký</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-2-p3s4a2">
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Gen</t_lemma>
       <functor>ADDR</functor>
       <tfa>t</tfa>
       <deepord>3</deepord>
      </LM>
      <LM id="t-ln94210-2-p3s4w5">
       <a>
        <lex.rf>a#a-ln94210-2-p3s4w5</lex.rf>
        <aux.rf>a#a-ln94210-2-p3s4w4</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>dokončení</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>after</subfunctor>
       <tfa>t</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>neut</gender>
        <number>sg</number>
        <negation>neg0</negation>
       </gram>
       <val_frame.rf>v#v-w597f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-2-p3s4w6">
         <a>
          <lex.rf>a#a-ln94210-2-p3s4w6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>projekt</t_lemma>
         <functor>PAT</functor>
         <tfa>t</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-2-p3s4a1">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>5</deepord>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-2-p3s4w8">
       <a>
        <lex.rf>a#a-ln94210-2-p3s4w8</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>výstavba</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>8</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-2-p3s4w9">
         <a>
          <lex.rf>a#a-ln94210-2-p3s4w9</lex.rf>
          <aux.rf>a#a-ln94210-2-p3s4w10</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>týkající_se</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>9</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
         <children>
          <LM id="t-ln94210-2-p3s4w12">
           <a>
            <lex.rf>a#a-ln94210-2-p3s4w12</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Evropa</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-2-p3s4w11">
             <a>
              <lex.rf>a#a-ln94210-2-p3s4w11</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>svobodný</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>11</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-39-p1s1">
   <atree.rf>a#a-ln94210-39-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-39-p1s1w3">
     <a>
      <lex.rf>a#a-ln94210-39-p1s1w3</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>chodit</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>4</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w1179f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-39-p1s1w2">
       <a>
        <lex.rf>a#a-ln94210-39-p1s1w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>člověk</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln94210-39-p1s1w1">
         <a>
          <lex.rf>a#a-ln94210-39-p1s1w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>mladý</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-39-p1s1a0">
       <nodetype>atom</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Neg</t_lemma>
       <functor>RHEM</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
      </LM>
      <LM id="t-ln94210-39-p1s1w5">
       <a>
        <lex.rf>a#a-ln94210-39-p1s1w5</lex.rf>
        <aux.rf>a#a-ln94210-39-p1s1w4</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>divadlo</t_lemma>
       <functor>DIR3</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>pl</number>
       </gram>
      </LM>
      <LM id="t-ln94210-39-p1s1w6">
       <a>
        <lex.rf>a#a-ln94210-39-p1s1w6</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>často</t_lemma>
       <functor>THO</functor>
       <tfa>f</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>adv.denot.grad.neg</sempos>
        <degcmp>pos</degcmp>
        <negation>neg0</negation>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-39-p2s1A">
   <atree.rf>a#a-ln94210-39-p2s1A</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-39-p2s1Aw1">
     <a>
      <lex.rf>a#a-ln94210-39-p2s1Aw1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>Praha</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>fem</gender>
      <number>sg</number>
     </gram>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-39-p2s1B">
   <atree.rf>a#a-ln94210-39-p2s1B</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-39-p2s1Bw5">
     <a>
      <lex.rf>a#a-ln94210-39-p2s1Bw5</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>patřit</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>5</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w3411f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-39-p2s1Bw1">
       <a>
        <lex.rf>a#a-ln94210-39-p2s1Bw1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>návštěva</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln94210-39-p2s1Bw3">
         <a>
          <lex.rf>a#a-ln94210-39-p2s1Bw3</lex.rf>
         </a>
         <nodetype>coap</nodetype>
         <t_lemma>a</t_lemma>
         <functor>CONJ</functor>
         <deepord>3</deepord>
         <children>
          <LM id="t-ln94210-39-p2s1Bw2">
           <a>
            <lex.rf>a#a-ln94210-39-p2s1Bw2</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>kino</t_lemma>
           <functor>PAT</functor>
           <is_member>1</is_member>
           <tfa>f</tfa>
           <deepord>2</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>pl</number>
           </gram>
          </LM>
          <LM id="t-ln94210-39-p2s1Bw4">
           <a>
            <lex.rf>a#a-ln94210-39-p2s1Bw4</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>divadlo</t_lemma>
           <functor>PAT</functor>
           <is_member>1</is_member>
           <tfa>f</tfa>
           <deepord>4</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>pl</number>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-39-p2s1Bw9">
       <a>
        <lex.rf>a#a-ln94210-39-p2s1Bw9</lex.rf>
        <aux.rf>a#a-ln94210-39-p2s1Bw6</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>aktivita</t_lemma>
       <functor>DIR3</functor>
       <subfunctor>betw</subfunctor>
       <tfa>f</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln94210-39-p2s1Bw8">
         <a>
          <lex.rf>a#a-ln94210-39-p2s1Bw8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>častý</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s1Bw7">
           <a>
            <lex.rf>a#a-ln94210-39-p2s1Bw7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>málo</t_lemma>
           <functor>EXT</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>adj.quant.grad</sempos>
            <degcmp>acomp</degcmp>
            <numertype>basic</numertype>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-39-p2s1Bw11">
         <a>
          <lex.rf>a#a-ln94210-39-p2s1Bw11</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>člověk</t_lemma>
         <functor>ACT</functor>
         <tfa>f</tfa>
         <deepord>9</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>anim</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s1Bw10">
           <a>
            <lex.rf>a#a-ln94210-39-p2s1Bw10</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>mladý</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
          <LM id="t-ln94210-39-p2s1Bw14">
           <a>
            <lex.rf>a#a-ln94210-39-p2s1Bw14</lex.rf>
            <aux.rf>a#a-ln94210-39-p2s1Bw12</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>republika</t_lemma>
           <functor>LOC</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>11</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-39-p2s1Bw13">
             <a>
              <lex.rf>a#a-ln94210-39-p2s1Bw13</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>Český</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>12</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-39-p2s2">
   <atree.rf>a#a-ln94210-39-p2s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-39-p2s2w1">
     <a>
      <lex.rf>a#a-ln94210-39-p2s2w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>vyplynout</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w8205f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-39-p2s2w2">
       <a>
        <lex.rf>a#a-ln94210-39-p2s2w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>ten</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.pron.def.demon</sempos>
        <gender>neut</gender>
        <number>sg</number>
       </gram>
       <coref_text.rf>t-ln94210-39-p2s1Bw5</coref_text.rf>
      </LM>
      <LM id="t-ln94210-39-p2s2w7">
       <a>
        <lex.rf>a#a-ln94210-39-p2s2w7</lex.rf>
        <aux.rf>a#a-ln94210-39-p2s2w3</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>výzkum</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-39-p2s2w8">
         <a>
          <lex.rf>a#a-ln94210-39-p2s2w8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Institut</t_lemma>
         <functor>ACT</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s2w10">
           <a>
            <lex.rf>a#a-ln94210-39-p2s2w10</lex.rf>
           </a>
           <nodetype>coap</nodetype>
           <t_lemma>a</t_lemma>
           <functor>CONJ</functor>
           <deepord>6</deepord>
           <children>
            <LM id="t-ln94210-39-p2s2w9">
             <a>
              <lex.rf>a#a-ln94210-39-p2s2w9</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>dítě</t_lemma>
             <functor>APP</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>5</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>pl</number>
             </gram>
            </LM>
            <LM id="t-ln94210-39-p2s2w11">
             <a>
              <lex.rf>a#a-ln94210-39-p2s2w11</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>mládež</t_lemma>
             <functor>APP</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>7</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-39-p2s2w12">
           <a>
            <lex.rf>a#a-ln94210-39-p2s2w12</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>ministerstvo</t_lemma>
           <functor>APP</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-39-p2s2w16">
             <a>
              <lex.rf>a#a-ln94210-39-p2s2w16</lex.rf>
             </a>
             <nodetype>coap</nodetype>
             <t_lemma>a</t_lemma>
             <functor>CONJ</functor>
             <deepord>11</deepord>
             <children>
              <LM id="t-ln94210-39-p2s2w13">
               <a>
                <lex.rf>a#a-ln94210-39-p2s2w13</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>školství</t_lemma>
               <functor>APP</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>9</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>neut</gender>
                <number>sg</number>
               </gram>
              </LM>
              <LM id="t-ln94210-39-p2s2w15">
               <a>
                <lex.rf>a#a-ln94210-39-p2s2w15</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>mládež</t_lemma>
               <functor>APP</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>10</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>sg</number>
               </gram>
              </LM>
              <LM id="t-ln94210-39-p2s2w17">
               <a>
                <lex.rf>a#a-ln94210-39-p2s2w17</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>tělovýchova</t_lemma>
               <functor>APP</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>12</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>sg</number>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-39-p2s2w5">
         <a>
          <lex.rf>a#a-ln94210-39-p2s2w5</lex.rf>
         </a>
         <nodetype>coap</nodetype>
         <t_lemma>a</t_lemma>
         <functor>CONJ</functor>
         <deepord>14</deepord>
         <children>
          <LM id="t-ln94210-39-p2s2w4">
           <a>
            <lex.rf>a#a-ln94210-39-p2s2w4</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>únorový</t_lemma>
           <functor>RSTR</functor>
           <is_member>1</is_member>
           <tfa>f</tfa>
           <deepord>13</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
          <LM id="t-ln94210-39-p2s2w6">
           <a>
            <lex.rf>a#a-ln94210-39-p2s2w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>květnový</t_lemma>
           <functor>RSTR</functor>
           <is_member>1</is_member>
           <tfa>f</tfa>
           <deepord>15</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-39-p2s2w20">
         <a>
          <lex.rf>a#a-ln94210-39-p2s2w20</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>pracovat</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>17</deepord>
         <gram>
          <sempos>v</sempos>
          <verbmod>ind</verbmod>
          <deontmod>decl</deontmod>
          <tense>ant</tense>
          <aspect>proc</aspect>
          <resultative>res0</resultative>
          <dispmod>disp0</dispmod>
          <iterativeness>it0</iterativeness>
         </gram>
         <val_frame.rf>v#v-w4243f6</val_frame.rf>
         <children>
          <LM id="t-ln94210-39-p2s2w19">
           <a>
            <lex.rf>a#a-ln94210-39-p2s2w19</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>který</t_lemma>
           <functor>ACT</functor>
           <tfa>t</tfa>
           <deepord>16</deepord>
           <gram>
            <sempos>n.pron.indef</sempos>
            <gender>inher</gender>
            <number>inher</number>
            <indeftype>relat</indeftype>
            <person>inher</person>
           </gram>
           <coref_gram.rf>t-ln94210-39-p2s2w7</coref_gram.rf>
          </LM>
          <LM id="t-ln94210-39-p2s2w23">
           <a>
            <lex.rf>a#a-ln94210-39-p2s2w23</lex.rf>
            <aux.rf>a#a-ln94210-39-p2s2w21</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>respondent</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>18</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>pl</number>
           </gram>
           <children>
            <LM id="t-ln94210-39-p2s2w22">
             <a>
              <lex.rf>a#a-ln94210-39-p2s2w22</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>758</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>19</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
            </LM>
            <LM id="t-ln94210-39-p2s2w25">
             <a>
              <lex.rf>a#a-ln94210-39-p2s2w25</lex.rf>
              <aux.rf>a#a-ln94210-39-p2s2w24</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>věk</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>20</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94210-39-p2s2w29">
               <a>
                <lex.rf>a#a-ln94210-39-p2s2w29</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>rok</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>21</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>neut</gender>
                <number>pl</number>
               </gram>
               <children>
                <LM id="t-ln94210-39-p2s2w27">
                 <a>
                  <lex.rf>a#a-ln94210-39-p2s2w27</lex.rf>
                 </a>
                 <nodetype>coap</nodetype>
                 <t_lemma>až</t_lemma>
                 <functor>OPER</functor>
                 <deepord>23</deepord>
                 <children>
                  <LM id="t-ln94210-39-p2s2w26">
                   <a>
                    <lex.rf>a#a-ln94210-39-p2s2w26</lex.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>15</t_lemma>
                   <functor>RSTR</functor>
                   <is_member>1</is_member>
                   <tfa>f</tfa>
                   <deepord>22</deepord>
                   <gram>
                    <sempos>adj.quant.def</sempos>
                    <numertype>basic</numertype>
                   </gram>
                  </LM>
                  <LM id="t-ln94210-39-p2s2w28">
                   <a>
                    <lex.rf>a#a-ln94210-39-p2s2w28</lex.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>29</t_lemma>
                   <functor>RSTR</functor>
                   <is_member>1</is_member>
                   <tfa>f</tfa>
                   <deepord>24</deepord>
                   <gram>
                    <sempos>adj.quant.def</sempos>
                    <numertype>basic</numertype>
                   </gram>
                  </LM>
                 </children>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-39-p2s3">
   <atree.rf>a#a-ln94210-39-p2s3</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-39-p2s3w9">
     <a>
      <lex.rf>a#a-ln94210-39-p2s3w9</lex.rf>
     </a>
     <nodetype>coap</nodetype>
     <t_lemma>#Comma</t_lemma>
     <functor>CONJ</functor>
     <deepord>9</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-ln94210-39-p2s3w4">
       <a>
        <lex.rf>a#a-ln94210-39-p2s3w4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>chodit</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>sim</tense>
        <aspect>proc</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w1179f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-39-p2s3w2">
         <a>
          <lex.rf>a#a-ln94210-39-p2s3w2</lex.rf>
          <aux.rf>a#a-ln94210-39-p2s3w1</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>divadlo</t_lemma>
         <functor>DIR3</functor>
         <subfunctor>basic</subfunctor>
         <tfa>c</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-39-p2s3a0">
         <nodetype>atom</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Neg</t_lemma>
         <functor>RHEM</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
        </LM>
        <LM id="t-ln94210-39-p2s3w3">
         <a>
          <lex.rf>a#a-ln94210-39-p2s3w3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>vůbec</t_lemma>
         <functor>EXT</functor>
         <subfunctor>basic</subfunctor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>adv.denot.ngrad.nneg</sempos>
         </gram>
        </LM>
        <LM id="t-ln94210-39-p2s3w6">
         <a>
          <lex.rf>a#a-ln94210-39-p2s3w6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>procento</t_lemma>
         <functor>ACT</functor>
         <tfa>f</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s3w8">
           <a>
            <lex.rf>a#a-ln94210-39-p2s3w8</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>člověk</t_lemma>
           <functor>MAT</functor>
           <tfa>t</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>pl</number>
           </gram>
           <children>
            <LM id="t-ln94210-39-p2s3w7">
             <a>
              <lex.rf>a#a-ln94210-39-p2s3w7</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>mladý</t_lemma>
             <functor>RSTR</functor>
             <tfa>t</tfa>
             <deepord>5</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-39-p2s3w5">
           <a>
            <lex.rf>a#a-ln94210-39-p2s3w5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>46</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-39-p2s3a1">
       <a>
        <lex.rf>a#a-ln94210-39-p2s3w4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>chodit</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>t</tfa>
       <deepord>12</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>sim</tense>
        <aspect>proc</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w1179f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-39-p2s3w11">
         <a>
          <lex.rf>a#a-ln94210-39-p2s3w11</lex.rf>
          <aux.rf>a#a-ln94210-39-p2s3w10</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>kino</t_lemma>
         <functor>DIR3</functor>
         <subfunctor>basic</subfunctor>
         <tfa>c</tfa>
         <deepord>10</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-39-p2s3a2">
         <nodetype>atom</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Neg</t_lemma>
         <functor>RHEM</functor>
         <tfa>t</tfa>
         <deepord>11</deepord>
        </LM>
        <LM id="t-ln94210-39-p2s3w13">
         <a>
          <lex.rf>a#a-ln94210-39-p2s3w13</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>procento</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>13</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s3w12">
           <a>
            <lex.rf>a#a-ln94210-39-p2s3w12</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>12</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>14</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-39-p2s4">
   <atree.rf>a#a-ln94210-39-p2s4</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-39-p2s4w14">
     <a>
      <lex.rf>a#a-ln94210-39-p2s4w14</lex.rf>
     </a>
     <nodetype>coap</nodetype>
     <t_lemma>#Comma</t_lemma>
     <functor>CONJ</functor>
     <deepord>12</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-ln94210-39-p2s4w6">
       <a>
        <lex.rf>a#a-ln94210-39-p2s4w6</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>navštívit</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>post</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w2319f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-39-p2s4w2">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w2</lex.rf>
          <aux.rf>a#a-ln94210-39-p2s4w1</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>ten</t_lemma>
         <functor>CPR</functor>
         <subfunctor>wrt</subfunctor>
         <tfa>c</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>n.pron.def.demon</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <coref_text.rf>t-ln94210-39-p2s3w9</coref_text.rf>
        </LM>
        <LM id="t-ln94210-39-p2s4w3">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w3</lex.rf>
         </a>
         <nodetype>atom</nodetype>
         <t_lemma>alespoň</t_lemma>
         <functor>RHEM</functor>
         <tfa>t</tfa>
         <deepord>2</deepord>
        </LM>
        <LM id="t-ln94210-39-p2s4w4">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w4</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>jeden</t_lemma>
         <functor>THO</functor>
         <tfa>c</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s4w5">
           <a>
            <lex.rf>a#a-ln94210-39-p2s4w5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>týdní</t_lemma>
           <functor>TWHEN</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>4</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-39-p2s4w7">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w7</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>divadlo</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-39-p2s4w9">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w9</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>procento</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>10</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s4w10">
           <a>
            <lex.rf>a#a-ln94210-39-p2s4w10</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>populace</t_lemma>
           <functor>MAT</functor>
           <tfa>t</tfa>
           <deepord>9</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-39-p2s4w13">
             <a>
              <lex.rf>a#a-ln94210-39-p2s4w13</lex.rf>
              <aux.rf>a#a-ln94210-39-p2s4w11</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>věk</t_lemma>
             <functor>RSTR</functor>
             <tfa>t</tfa>
             <deepord>8</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94210-39-p2s4w12">
               <a>
                <lex.rf>a#a-ln94210-39-p2s4w12</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>tento</t_lemma>
               <functor>RSTR</functor>
               <tfa>t</tfa>
               <deepord>7</deepord>
               <gram>
                <sempos>adj.pron.def.demon</sempos>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-39-p2s4w8">
           <a>
            <lex.rf>a#a-ln94210-39-p2s4w8</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>1</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>11</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-39-p2s4w20">
       <a>
        <lex.rf>a#a-ln94210-39-p2s4w20</lex.rf>
        <aux.rf>a#a-ln94210-39-p2s4w17</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>vypravit_se</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>16</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>post</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w8248f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-39-p2s4w16">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w16</lex.rf>
          <aux.rf>a#a-ln94210-39-p2s4w15</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>kino</t_lemma>
         <functor>DIR3</functor>
         <subfunctor>basic</subfunctor>
         <tfa>c</tfa>
         <deepord>13</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-39-p2s4w19">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w19</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>často</t_lemma>
         <functor>THO</functor>
         <tfa>t</tfa>
         <deepord>15</deepord>
         <gram>
          <sempos>adv.denot.grad.neg</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s4w18">
           <a>
            <lex.rf>a#a-ln94210-39-p2s4w18</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>takto</t_lemma>
           <functor>MANN</functor>
           <tfa>t</tfa>
           <deepord>14</deepord>
           <gram>
            <sempos>adv.denot.ngrad.nneg</sempos>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-39-p2s4w22">
         <a>
          <lex.rf>a#a-ln94210-39-p2s4w22</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>mladý</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>17</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>anim</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln94210-39-p2s4w21">
           <a>
            <lex.rf>a#a-ln94210-39-p2s4w21</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>17</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>18</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
          <LM id="t-ln94210-39-p2s4w24">
           <a>
            <lex.rf>a#a-ln94210-39-p2s4w24</lex.rf>
            <aux.rf>a#a-ln94210-39-p2s4w23</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>sto</t_lemma>
           <functor>DIR1</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>19</deepord>
           <gram>
            <sempos>n.quant.def</sempos>
            <gender>neut</gender>
            <number>sg</number>
            <numertype>basic</numertype>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-mf920925-022-p1s1">
   <atree.rf>a#a-mf920925-022-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-mf920925-022-p1s1w1">
     <a>
      <lex.rf>a#a-mf920925-022-p1s1w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>tip</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>inan</gender>
      <number>pl</number>
     </gram>
     <children>
      <LM id="t-mf920925-022-p1s1w2">
       <a>
        <lex.rf>a#a-mf920925-022-p1s1w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>dnes</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>adv.denot.ngrad.nneg</sempos>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-mf920925-022-p2s1">
   <atree.rf>a#a-mf920925-022-p2s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-mf920925-022-p2s1w1">
     <a>
      <lex.rf>a#a-mf920925-022-p2s1w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>Ruby</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>nr</gender>
      <number>nr</number>
     </gram>
    </LM>
   </children>
  </LM>
  <LM id="t-mf920925-022-p2s2">
   <atree.rf>a#a-mf920925-022-p2s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-mf920925-022-p2s2w27">
     <a>
      <lex.rf>a#a-mf920925-022-p2s2w27</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>uvádět</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>21</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w7420f3</val_frame.rf>
     <children>
      <LM id="t-mf920925-022-p2s2w7">
       <a>
        <lex.rf>a#a-mf920925-022-p2s2w7</lex.rf>
       </a>
       <nodetype>coap</nodetype>
       <t_lemma>#Comma</t_lemma>
       <functor>APPS</functor>
       <deepord>7</deepord>
       <children>
        <LM id="t-mf920925-022-p2s2w3">
         <a>
          <lex.rf>a#a-mf920925-022-p2s2w3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>drama</t_lemma>
         <functor>PAT</functor>
         <is_member>1</is_member>
         <tfa>c</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-mf920925-022-p2s2w2">
           <a>
            <lex.rf>a#a-mf920925-022-p2s2w2</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>filmový</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>2</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
          <LM id="t-mf920925-022-p2s2w1">
           <a>
            <lex.rf>a#a-mf920925-022-p2s2w1</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>americký</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>3</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
          <LM id="t-mf920925-022-p2s2w6">
           <a>
            <lex.rf>a#a-mf920925-022-p2s2w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Mackenzie</t_lemma>
           <functor>AUTH</functor>
           <tfa>f</tfa>
           <deepord>4</deepord>
           <is_name_of_person>1</is_name_of_person>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-mf920925-022-p2s2w5">
             <a>
              <lex.rf>a#a-mf920925-022-p2s2w5</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>John</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>5</deepord>
             <is_name_of_person>1</is_name_of_person>
             <gram>
              <sempos>n.denot</sempos>
              <gender>anim</gender>
              <number>sg</number>
             </gram>
            </LM>
            <LM id="t-mf920925-022-p2s2w4">
             <a>
              <lex.rf>a#a-mf920925-022-p2s2w4</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>režisér</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>6</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>anim</gender>
              <number>sg</number>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-mf920925-022-p2s2w15">
         <a>
          <lex.rf>a#a-mf920925-022-p2s2w15</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>fikce</t_lemma>
         <functor>PAT</functor>
         <is_member>1</is_member>
         <tfa>c</tfa>
         <deepord>13</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-mf920925-022-p2s2w11">
           <a>
            <lex.rf>a#a-mf920925-022-p2s2w11</lex.rf>
            <aux.rf>a#a-mf920925-022-p2s2w8</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>rekonstrukce</t_lemma>
           <functor>TWHEN</functor>
           <subfunctor>after</subfunctor>
           <tfa>t</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-mf920925-022-p2s2w10">
             <a>
              <lex.rf>a#a-mf920925-022-p2s2w10</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>paradokumentární</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>9</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
            <LM id="t-mf920925-022-p2s2w9">
             <a>
              <lex.rf>a#a-mf920925-022-p2s2w9</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>Stoneův</t_lemma>
             <functor>AUTH</functor>
             <tfa>f</tfa>
             <deepord>10</deepord>
             <is_name_of_person>1</is_name_of_person>
             <gram>
              <sempos>n.denot</sempos>
              <gender>anim</gender>
              <number>sg</number>
             </gram>
            </LM>
            <LM id="t-mf920925-022-p2s2w12">
             <a>
              <lex.rf>a#a-mf920925-022-p2s2w12</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>jméno</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>11</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>neut</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-mf920925-022-p2s2w13">
               <a>
                <lex.rf>a#a-mf920925-022-p2s2w13</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>JFK</t_lemma>
               <functor>ID</functor>
               <tfa>f</tfa>
               <deepord>12</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>anim</gender>
                <number>nr</number>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
          <LM id="t-mf920925-022-p2s2w18">
           <a>
            <lex.rf>a#a-mf920925-022-p2s2w18</lex.rf>
            <aux.rf>a#a-mf920925-022-p2s2w16</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>pozadí</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>14</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-mf920925-022-p2s2w19">
             <a>
              <lex.rf>a#a-mf920925-022-p2s2w19</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>atentát</t_lemma>
             <functor>APP</functor>
             <tfa>f</tfa>
             <deepord>15</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-mf920925-022-p2s2w25">
               <a>
                <lex.rf>a#a-mf920925-022-p2s2w25</lex.rf>
                <aux.rf>a#a-mf920925-022-p2s2w20</aux.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>Kennedy</t_lemma>
               <functor>PAT</functor>
               <tfa>f</tfa>
               <deepord>16</deepord>
               <is_name_of_person>1</is_name_of_person>
               <gram>
                <sempos>n.denot</sempos>
                <gender>anim</gender>
                <number>sg</number>
               </gram>
               <children>
                <LM id="t-mf920925-022-p2s2w21">
                 <a>
                  <lex.rf>a#a-mf920925-022-p2s2w21</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>J</t_lemma>
                 <functor>RSTR</functor>
                 <tfa>f</tfa>
                 <deepord>17</deepord>
                 <is_name_of_person>1</is_name_of_person>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>anim</gender>
                  <number>nr</number>
                 </gram>
                </LM>
                <LM id="t-mf920925-022-p2s2w23">
                 <a>
                  <lex.rf>a#a-mf920925-022-p2s2w23</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>F</t_lemma>
                 <functor>RSTR</functor>
                 <tfa>f</tfa>
                 <deepord>18</deepord>
                 <is_name_of_person>1</is_name_of_person>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>anim</gender>
                  <number>nr</number>
                 </gram>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
            <LM id="t-mf920925-022-p2s2w17">
             <a>
              <lex.rf>a#a-mf920925-022-p2s2w17</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>možný</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>19</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-mf920925-022-p2s2w14">
           <a>
            <lex.rf>a#a-mf920925-022-p2s2w14</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>další</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>20</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-mf920925-022-p2s2w29">
       <a>
        <lex.rf>a#a-mf920925-022-p2s2w29</lex.rf>
        <aux.rf>a#a-mf920925-022-p2s2w28</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>včerejšek</t_lemma>
       <functor>TSIN</functor>
       <tfa>f</tfa>
       <deepord>22</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-mf920925-022-p2s2w31">
       <a>
        <lex.rf>a#a-mf920925-022-p2s2w31</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>kino</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>23</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-mf920925-022-p2s2w30">
         <a>
          <lex.rf>a#a-mf920925-022-p2s2w30</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>pražský</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>24</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-mf920925-022-p2s2w32">
         <a>
          <lex.rf>a#a-mf920925-022-p2s2w32</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Kotva</t_lemma>
         <functor>ID</functor>
         <tfa>f</tfa>
         <deepord>25</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-mf920925-022-p3s1">
   <atree.rf>a#a-mf920925-022-p3s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-mf920925-022-p3s1w3">
     <a>
      <lex.rf>a#a-mf920925-022-p3s1w3</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>Brandauer</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <is_name_of_person>1</is_name_of_person>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>anim</gender>
      <number>sg</number>
     </gram>
     <children>
      <LM id="t-mf920925-022-p3s1w1">
       <a>
        <lex.rf>a#a-mf920925-022-p3s1w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Klaus</t_lemma>
       <functor>RSTR</functor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <is_name_of_person>1</is_name_of_person>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-mf920925-022-p3s1w2">
       <a>
        <lex.rf>a#a-mf920925-022-p3s1w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Maria</t_lemma>
       <functor>RSTR</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <is_name_of_person>1</is_name_of_person>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>sg</number>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-mf920925-022-p3s2">
   <atree.rf>a#a-mf920925-022-p3s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-mf920925-022-p3s2w20">
     <a>
      <lex.rf>a#a-mf920925-022-p3s2w20</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>vystoupit</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>14</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>post</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w8396f3</val_frame.rf>
     <children>
      <LM id="t-mf920925-022-p3s2w3">
       <a>
        <lex.rf>a#a-mf920925-022-p3s2w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>herec</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-mf920925-022-p3s2w2">
         <a>
          <lex.rf>a#a-mf920925-022-p3s2w2</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>rakouský</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-mf920925-022-p3s2w1">
         <a>
          <lex.rf>a#a-mf920925-022-p3s2w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>populární</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-mf920925-022-p3s2w7">
         <a>
          <lex.rf>a#a-mf920925-022-p3s2w7</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>známý</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
         <children>
          <LM id="t-mf920925-022-p3s2w6">
           <a>
            <lex.rf>a#a-mf920925-022-p3s2w6</lex.rf>
            <aux.rf>a#a-mf920925-022-p3s2w5</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>#PersPron</t_lemma>
           <functor>LOC</functor>
           <subfunctor>near</subfunctor>
           <tfa>t</tfa>
           <deepord>4</deepord>
           <gram>
            <sempos>n.pron.def.pers</sempos>
            <gender>nr</gender>
            <number>pl</number>
            <person>1</person>
            <politeness>basic</politeness>
           </gram>
          </LM>
          <LM id="t-mf920925-022-p3s2w8">
           <a>
            <lex.rf>a#a-mf920925-022-p3s2w8</lex.rf>
           </a>
           <nodetype>atom</nodetype>
           <t_lemma>zejména</t_lemma>
           <functor>RHEM</functor>
           <tfa>f</tfa>
           <deepord>6</deepord>
          </LM>
          <LM id="t-mf920925-022-p3s2w10">
           <a>
            <lex.rf>a#a-mf920925-022-p3s2w10</lex.rf>
            <aux.rf>a#a-mf920925-022-p3s2w9</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>film</t_lemma>
           <functor>DIR1</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>pl</number>
           </gram>
           <children>
            <LM id="t-mf920925-022-p3s2w15">
             <a>
              <lex.rf>a#a-mf920925-022-p3s2w15</lex.rf>
             </a>
             <nodetype>coap</nodetype>
             <t_lemma>a</t_lemma>
             <functor>CONJ</functor>
             <deepord>11</deepord>
             <children>
              <LM id="t-mf920925-022-p3s2w11">
               <a>
                <lex.rf>a#a-mf920925-022-p3s2w11</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>Mefisto</t_lemma>
               <functor>ID</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>8</deepord>
               <is_name_of_person>1</is_name_of_person>
               <gram>
                <sempos>n.denot</sempos>
                <gender>anim</gender>
                <number>sg</number>
               </gram>
              </LM>
              <LM id="t-mf920925-022-p3s2w14">
               <a>
                <lex.rf>a#a-mf920925-022-p3s2w14</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>Redl</t_lemma>
               <functor>ID</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>9</deepord>
               <is_name_of_person>1</is_name_of_person>
               <gram>
                <sempos>n.denot</sempos>
                <gender>anim</gender>
                <number>sg</number>
               </gram>
               <children>
                <LM id="t-mf920925-022-p3s2w13">
                 <a>
                  <lex.rf>a#a-mf920925-022-p3s2w13</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>Plukovník</t_lemma>
                 <functor>RSTR</functor>
                 <tfa>f</tfa>
                 <deepord>10</deepord>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>anim</gender>
                  <number>sg</number>
                 </gram>
                </LM>
               </children>
              </LM>
              <LM id="t-mf920925-022-p3s2w16">
               <a>
                <lex.rf>a#a-mf920925-022-p3s2w16</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>Vzpomínka</t_lemma>
               <functor>ID</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>12</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>pl</number>
               </gram>
               <children>
                <LM id="t-mf920925-022-p3s2w18">
                 <a>
                  <lex.rf>a#a-mf920925-022-p3s2w18</lex.rf>
                  <aux.rf>a#a-mf920925-022-p3s2w17</aux.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>Afrika</t_lemma>
                 <functor>PAT</functor>
                 <tfa>f</tfa>
                 <deepord>13</deepord>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>fem</gender>
                  <number>sg</number>
                 </gram>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-mf920925-022-p3s2w22">
       <a>
        <lex.rf>a#a-mf920925-022-p3s2w22</lex.rf>
        <aux.rf>a#a-mf920925-022-p3s2w21</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>neděle</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>15</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-mf920925-022-p3s2w25">
       <a>
        <lex.rf>a#a-mf920925-022-p3s2w25</lex.rf>
        <aux.rf>a#a-mf920925-022-p3s2w23</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>hodina</t_lemma>
       <functor>TSIN</functor>
       <tfa>t</tfa>
       <deepord>16</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>nr</number>
       </gram>
       <children>
        <LM id="t-mf920925-022-p3s2w24">
         <a>
          <lex.rf>a#a-mf920925-022-p3s2w24</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>11.00</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>17</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-mf920925-022-p3s2w28">
       <a>
        <lex.rf>a#a-mf920925-022-p3s2w28</lex.rf>
        <aux.rf>a#a-mf920925-022-p3s2w26</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Studio</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>18</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-mf920925-022-p3s2w27">
         <a>
          <lex.rf>a#a-mf920925-022-p3s2w27</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>pražský</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>19</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-mf920925-022-p3s2w29">
         <a>
          <lex.rf>a#a-mf920925-022-p3s2w29</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Labyrint</t_lemma>
         <functor>ID</functor>
         <tfa>f</tfa>
         <deepord>20</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-mf920925-022-p3s2w31">
       <a>
        <lex.rf>a#a-mf920925-022-p3s2w31</lex.rf>
        <aux.rf>a#a-mf920925-022-p3s2w30</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>čtení</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>23</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>neut</gender>
        <number>sg</number>
        <negation>neg0</negation>
       </gram>
       <val_frame.rf>v#v-w353f1</val_frame.rf>
       <children>
        <LM id="t-mf920925-022-p3s2a0">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>21</deepord>
        </LM>
        <LM id="t-mf920925-022-p3s2a1">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>PAT</functor>
         <tfa>t</tfa>
         <deepord>22</deepord>
        </LM>
        <LM id="t-mf920925-022-p3s2w33">
         <a>
          <lex.rf>a#a-mf920925-022-p3s2w33</lex.rf>
          <aux.rf>a#a-mf920925-022-p3s2w32</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>dílo</t_lemma>
         <functor>DIR1</functor>
         <subfunctor>basic</subfunctor>
         <tfa>f</tfa>
         <deepord>24</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-mf920925-022-p3s2w35">
           <a>
            <lex.rf>a#a-mf920925-022-p3s2w35</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Torberg</t_lemma>
           <functor>AUTH</functor>
           <tfa>f</tfa>
           <deepord>25</deepord>
           <is_name_of_person>1</is_name_of_person>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-mf920925-022-p3s2w34">
             <a>
              <lex.rf>a#a-mf920925-022-p3s2w34</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>Friedrich</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>26</deepord>
             <is_name_of_person>1</is_name_of_person>
             <gram>
              <sempos>n.denot</sempos>
              <gender>anim</gender>
              <number>sg</number>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-mf920925-022-p4s1A">
   <atree.rf>a#a-mf920925-022-p4s1A</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-mf920925-022-p4s1Aa0">
     <nodetype>qcomplex</nodetype>
     <is_generated>1</is_generated>
     <t_lemma>#EmpVerb</t_lemma>
     <functor>PRED</functor>
     <tfa>t</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-mf920925-022-p4s1Aa1">
       <a>
        <lex.rf>a#a-mf920925-022-p4s1Aw3</lex.rf>
       </a>
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Unsp</t_lemma>
       <functor>ACMP</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>1</deepord>
      </LM>
      <LM id="t-mf920925-022-p4s1Aw2">
       <a>
        <lex.rf>a#a-mf920925-022-p4s1Aw2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>minuta</t_lemma>
       <functor>THL</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-mf920925-022-p4s1Aw1">
         <a>
          <lex.rf>a#a-mf920925-022-p4s1Aw1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>deset</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-19-p1s1">
   <atree.rf>a#a-ln94210-19-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-19-p1s1w2">
     <a>
      <lex.rf>a#a-ln94210-19-p1s1w2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>přesedlávat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w4960f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-19-p1s1w1">
       <a>
        <lex.rf>a#a-ln94210-19-p1s1w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Čína</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln94210-19-p1s1w4">
       <a>
        <lex.rf>a#a-ln94210-19-p1s1w4</lex.rf>
        <aux.rf>a#a-ln94210-19-p1s1w3</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>kolo</t_lemma>
       <functor>DIR1</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>pl</number>
       </gram>
      </LM>
      <LM id="t-ln94210-19-p1s1w6">
       <a>
        <lex.rf>a#a-ln94210-19-p1s1w6</lex.rf>
        <aux.rf>a#a-ln94210-19-p1s1w5</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>auto</t_lemma>
       <functor>DIR3</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>4</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>pl</number>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-19-p2s1A">
   <atree.rf>a#a-ln94210-19-p2s1A</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-19-p2s1Aw1">
     <a>
      <lex.rf>a#a-ln94210-19-p2s1Aw1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>Peking</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>inan</gender>
      <number>sg</number>
     </gram>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-19-p2s1B">
   <atree.rf>a#a-ln94210-19-p2s1B</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-19-p2s1Bw6">
     <a>
      <lex.rf>a#a-ln94210-19-p2s1Bw6</lex.rf>
     </a>
     <nodetype>coap</nodetype>
     <t_lemma>#Colon</t_lemma>
     <functor>APPS</functor>
     <deepord>5</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-ln94210-19-p2s1Bw2">
       <a>
        <lex.rf>a#a-ln94210-19-p2s1Bw2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>sázet</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>sim</tense>
        <aspect>proc</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w5898f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-19-p2s1Bw1">
         <a>
          <lex.rf>a#a-ln94210-19-p2s1Bw1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Čína</t_lemma>
         <functor>ACT</functor>
         <tfa>c</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-19-p2s1Bw4">
         <a>
          <lex.rf>a#a-ln94210-19-p2s1Bw4</lex.rf>
          <aux.rf>a#a-ln94210-19-p2s1Bw3</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>výroba</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-19-p2s1Bw5">
           <a>
            <lex.rf>a#a-ln94210-19-p2s1Bw5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>automobil</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>4</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>pl</number>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-19-p2s1Bw42">
       <a>
        <lex.rf>a#a-ln94210-19-p2s1Bw42</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>uvést</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>8</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>ant</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w7449f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-19-p2s1Bw43">
         <a>
          <lex.rf>a#a-ln94210-19-p2s1Bw43</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>agentura</t_lemma>
         <functor>ACT</functor>
         <tfa>c</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-19-p2s1Bw44">
           <a>
            <lex.rf>a#a-ln94210-19-p2s1Bw44</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>DPA</t_lemma>
           <functor>ID</functor>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>nr</number>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-19-p2s1Bw23">
         <a>
          <lex.rf>a#a-ln94210-19-p2s1Bw23</lex.rf>
          <aux.rf>a#a-ln94210-19-p2s1Bw13</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>poskytnout</t_lemma>
         <functor>EFF</functor>
         <tfa>f</tfa>
         <deepord>25</deepord>
         <is_dsp_root>1</is_dsp_root>
         <sentmod>enunc</sentmod>
         <gram>
          <sempos>v</sempos>
          <verbmod>ind</verbmod>
          <deontmod>decl</deontmod>
          <tense>ant</tense>
          <aspect>cpl</aspect>
          <resultative>res0</resultative>
          <dispmod>disp0</dispmod>
          <iterativeness>it0</iterativeness>
         </gram>
         <val_frame.rf>v#v-w3973f2</val_frame.rf>
         <children>
          <LM id="t-ln94210-19-p2s1Bw10">
           <a>
            <lex.rf>a#a-ln94210-19-p2s1Bw10</lex.rf>
            <aux.rf>a#a-ln94210-19-p2s1Bw7</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>země</t_lemma>
           <functor>LOC</functor>
           <subfunctor>basic</subfunctor>
           <tfa>t</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-19-p2s1Bw8">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw8</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>tento</t_lemma>
             <functor>RSTR</functor>
             <tfa>t</tfa>
             <deepord>9</deepord>
             <gram>
              <sempos>adj.pron.def.demon</sempos>
             </gram>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw12">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw12</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>kolo</t_lemma>
             <functor>APP</functor>
             <tfa>f</tfa>
             <deepord>11</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>neut</gender>
              <number>pl</number>
             </gram>
             <children>
              <LM id="t-ln94210-19-p2s1Bw11">
               <a>
                <lex.rf>a#a-ln94210-19-p2s1Bw11</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>jízdní</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>12</deepord>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>pos</degcmp>
                <negation>neg0</negation>
               </gram>
              </LM>
             </children>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw9">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw9</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>klasický</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>13</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-19-p2s1Bw16">
           <a>
            <lex.rf>a#a-ln94210-19-p2s1Bw16</lex.rf>
            <aux.rf>a#a-ln94210-19-p2s1Bw14</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>vydání</t_lemma>
           <functor>CRIT</functor>
           <tfa>t</tfa>
           <deepord>15</deepord>
           <gram>
            <sempos>n.denot.neg</sempos>
            <gender>neut</gender>
            <number>sg</number>
            <negation>neg0</negation>
           </gram>
           <val_frame.rf>v#v-w7839f1</val_frame.rf>
           <children>
            <LM id="t-ln94210-19-p2s1Ba2">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#Gen</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>14</deepord>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw17">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw17</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>list</t_lemma>
             <functor>PAT</functor>
             <tfa>f</tfa>
             <deepord>16</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94210-19-p2s1Ba3">
               <nodetype>list</nodetype>
               <is_generated>1</is_generated>
               <t_lemma>#Forn</t_lemma>
               <functor>ID</functor>
               <tfa>f</tfa>
               <deepord>18</deepord>
               <children>
                <LM id="t-ln94210-19-p2s1Bw18">
                 <a>
                  <lex.rf>a#a-ln94210-19-p2s1Bw18</lex.rf>
                 </a>
                 <nodetype>fphr</nodetype>
                 <t_lemma>Business</t_lemma>
                 <functor>FPHR</functor>
                 <deepord>17</deepord>
                </LM>
                <LM id="t-ln94210-19-p2s1Bw19">
                 <a>
                  <lex.rf>a#a-ln94210-19-p2s1Bw19</lex.rf>
                 </a>
                 <nodetype>fphr</nodetype>
                 <t_lemma>Weekly</t_lemma>
                 <functor>FPHR</functor>
                 <deepord>19</deepord>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw15">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw15</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>poslední</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>20</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-19-p2s1Bw22">
           <a>
            <lex.rf>a#a-ln94210-19-p2s1Bw22</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>automobilka</t_lemma>
           <functor>ADDR</functor>
           <tfa>c</tfa>
           <deepord>22</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94210-19-p2s1Bw21">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw21</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>domácí</t_lemma>
             <functor>RSTR</functor>
             <tfa>t</tfa>
             <deepord>21</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw20">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw20</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>vedoucí</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>23</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-19-p2s1Ba1">
           <nodetype>qcomplex</nodetype>
           <is_generated>1</is_generated>
           <t_lemma>#Gen</t_lemma>
           <functor>ACT</functor>
           <tfa>t</tfa>
           <deepord>24</deepord>
          </LM>
          <LM id="t-ln94210-19-p2s1Bw25">
           <a>
            <lex.rf>a#a-ln94210-19-p2s1Bw25</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>dotace</t_lemma>
           <functor>CPHR</functor>
           <tfa>f</tfa>
           <deepord>28</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <val_frame.rf>v#v-w749f1</val_frame.rf>
           <children>
            <LM id="t-ln94210-19-p2s1Ba4">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#QCor</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>26</deepord>
             <coref_gram.rf>t-ln94210-19-p2s1Ba1</coref_gram.rf>
            </LM>
            <LM id="t-ln94210-19-p2s1Ba5">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#QCor</t_lemma>
             <functor>ADDR</functor>
             <tfa>t</tfa>
             <deepord>27</deepord>
             <coref_gram.rf>t-ln94210-19-p2s1Bw22</coref_gram.rf>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw24">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw24</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>vládní</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>29</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw27">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw27</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>milión</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>30</deepord>
             <gram>
              <sempos>n.quant.def</sempos>
              <gender>inan</gender>
              <number>pl</number>
              <numertype>basic</numertype>
             </gram>
             <children>
              <LM id="t-ln94210-19-p2s1Bw28">
               <a>
                <lex.rf>a#a-ln94210-19-p2s1Bw28</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>USD</t_lemma>
               <functor>MAT</functor>
               <tfa>f</tfa>
               <deepord>31</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>inan</gender>
                <number>nr</number>
               </gram>
              </LM>
              <LM id="t-ln94210-19-p2s1Bw26">
               <a>
                <lex.rf>a#a-ln94210-19-p2s1Bw26</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>250</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>32</deepord>
               <gram>
                <sempos>adj.quant.def</sempos>
                <numertype>basic</numertype>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94210-19-p2s1Bw34">
           <a>
            <lex.rf>a#a-ln94210-19-p2s1Bw34</lex.rf>
            <aux.rf>a#a-ln94210-19-p2s1Bw30</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>hlásit</t_lemma>
           <functor>CONTRD</functor>
           <tfa>f</tfa>
           <deepord>36</deepord>
           <gram>
            <sempos>v</sempos>
            <verbmod>ind</verbmod>
            <deontmod>decl</deontmod>
            <tense>sim</tense>
            <aspect>proc</aspect>
            <resultative>res0</resultative>
            <dispmod>disp0</dispmod>
            <iterativeness>it0</iterativeness>
           </gram>
           <val_frame.rf>v#v-w1050f1</val_frame.rf>
           <children>
            <LM id="t-ln94210-19-p2s1Bw31">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw31</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>výrobce</t_lemma>
             <functor>ACT</functor>
             <tfa>c</tfa>
             <deepord>33</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>anim</gender>
              <number>pl</number>
             </gram>
             <children>
              <LM id="t-ln94210-19-p2s1Bw33">
               <a>
                <lex.rf>a#a-ln94210-19-p2s1Bw33</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>kolo</t_lemma>
               <functor>PAT</functor>
               <tfa>f</tfa>
               <deepord>34</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>neut</gender>
                <number>pl</number>
               </gram>
               <children>
                <LM id="t-ln94210-19-p2s1Bw32">
                 <a>
                  <lex.rf>a#a-ln94210-19-p2s1Bw32</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>jízdní</t_lemma>
                 <functor>RSTR</functor>
                 <tfa>f</tfa>
                 <deepord>35</deepord>
                 <gram>
                  <sempos>adj.denot</sempos>
                  <degcmp>pos</degcmp>
                  <negation>neg0</negation>
                 </gram>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw37">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw37</lex.rf>
              <aux.rf>a#a-ln94210-19-p2s1Bw35</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>pololetí</t_lemma>
             <functor>THL</functor>
             <tfa>f</tfa>
             <deepord>37</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>neut</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94210-19-p2s1Bw36">
               <a>
                <lex.rf>a#a-ln94210-19-p2s1Bw36</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>jeden</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>38</deepord>
               <gram>
                <sempos>adj.quant.def</sempos>
                <numertype>ord</numertype>
               </gram>
              </LM>
             </children>
            </LM>
            <LM id="t-ln94210-19-p2s1Bw39">
             <a>
              <lex.rf>a#a-ln94210-19-p2s1Bw39</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>pokles</t_lemma>
             <functor>PAT</functor>
             <tfa>f</tfa>
             <deepord>39</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94210-19-p2s1Bw40">
               <a>
                <lex.rf>a#a-ln94210-19-p2s1Bw40</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>odbyt</t_lemma>
               <functor>ACT</functor>
               <tfa>f</tfa>
               <deepord>40</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>inan</gender>
                <number>sg</number>
               </gram>
              </LM>
              <LM id="t-ln94210-19-p2s1Bw38">
               <a>
                <lex.rf>a#a-ln94210-19-p2s1Bw38</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>podstatný</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>41</deepord>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>pos</degcmp>
                <negation>neg0</negation>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-19-p3s1">
   <atree.rf>a#a-ln94210-19-p3s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-19-p3s1w5">
     <a>
      <lex.rf>a#a-ln94210-19-p3s1w5</lex.rf>
      <aux.rf>a#a-ln94210-19-p3s1w4</aux.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>poskytnout</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>5</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w3973f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-19-p3s1w3">
       <a>
        <lex.rf>a#a-ln94210-19-p3s1w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>dotace</t_lemma>
       <functor>PAT</functor>
       <tfa>t</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-19-p3s1w1">
         <a>
          <lex.rf>a#a-ln94210-19-p3s1w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>zmíněný</t_lemma>
         <functor>RSTR</functor>
         <tfa>t</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-ln94210-19-p3s1w2">
         <a>
          <lex.rf>a#a-ln94210-19-p3s1w2</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>čtvrtmiliardový</t_lemma>
         <functor>RSTR</functor>
         <tfa>t</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-19-p3s1a0">
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Gen</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>4</deepord>
      </LM>
      <LM id="t-ln94210-19-p3s1w6">
       <a>
        <lex.rf>a#a-ln94210-19-p3s1w6</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>automobilka</t_lemma>
       <functor>ADDR</functor>
       <tfa>t</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-19-p3s1a1">
         <nodetype>list</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Forn</t_lemma>
         <functor>ID</functor>
         <tfa>f</tfa>
         <deepord>10</deepord>
         <children>
          <LM id="t-ln94210-19-p3s1w7">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w7</lex.rf>
           </a>
           <nodetype>fphr</nodetype>
           <t_lemma>Tianjin</t_lemma>
           <functor>FPHR</functor>
           <deepord>7</deepord>
          </LM>
          <LM id="t-ln94210-19-p3s1w8">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w8</lex.rf>
           </a>
           <nodetype>fphr</nodetype>
           <t_lemma>Automotive</t_lemma>
           <functor>FPHR</functor>
           <deepord>8</deepord>
          </LM>
          <LM id="t-ln94210-19-p3s1w9">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w9</lex.rf>
           </a>
           <nodetype>fphr</nodetype>
           <t_lemma>Industry</t_lemma>
           <functor>FPHR</functor>
           <deepord>9</deepord>
          </LM>
          <LM id="t-ln94210-19-p3s1w10">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w10</lex.rf>
           </a>
           <nodetype>fphr</nodetype>
           <t_lemma>Corp</t_lemma>
           <functor>FPHR</functor>
           <deepord>11</deepord>
          </LM>
          <LM id="t-ln94210-19-p3s1w11">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w11</lex.rf>
           </a>
           <nodetype>fphr</nodetype>
           <t_lemma>#Period</t_lemma>
           <functor>FPHR</functor>
           <deepord>12</deepord>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94210-19-p3s1w13">
       <a>
        <lex.rf>a#a-ln94210-19-p3s1w13</lex.rf>
        <aux.rf>a#a-ln94210-19-p3s1w12</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>zvýšení</t_lemma>
       <functor>AIM</functor>
       <tfa>f</tfa>
       <deepord>14</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>neut</gender>
        <number>sg</number>
        <negation>neg0</negation>
       </gram>
       <val_frame.rf>v#v-w10004f1</val_frame.rf>
       <children>
        <LM id="t-ln94210-19-p3s1a2">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>13</deepord>
        </LM>
        <LM id="t-ln94210-19-p3s1w14">
         <a>
          <lex.rf>a#a-ln94210-19-p3s1w14</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>výroba</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>15</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94210-19-p3s1w18">
         <a>
          <lex.rf>a#a-ln94210-19-p3s1w18</lex.rf>
          <aux.rf>a#a-ln94210-19-p3s1w15</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>tisíc</t_lemma>
         <functor>ORIG</functor>
         <tfa>f</tfa>
         <deepord>16</deepord>
         <gram>
          <sempos>n.quant.def</sempos>
          <gender>inan</gender>
          <number>sg</number>
          <numertype>basic</numertype>
         </gram>
         <children>
          <LM id="t-ln94210-19-p3s1w17">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w17</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>60</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>17</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
          <LM id="t-ln94210-19-p3s1w16">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w16</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>současný</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>18</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-19-p3s1w21">
         <a>
          <lex.rf>a#a-ln94210-19-p3s1w21</lex.rf>
          <aux.rf>a#a-ln94210-19-p3s1w19</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>tisíc</t_lemma>
         <functor>EFF</functor>
         <tfa>f</tfa>
         <deepord>19</deepord>
         <gram>
          <sempos>n.quant.def</sempos>
          <gender>inan</gender>
          <number>sg</number>
          <numertype>basic</numertype>
         </gram>
         <children>
          <LM id="t-ln94210-19-p3s1w22">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w22</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>vůz</t_lemma>
           <functor>MAT</functor>
           <tfa>f</tfa>
           <deepord>20</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>pl</number>
           </gram>
          </LM>
          <LM id="t-ln94210-19-p3s1w20">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w20</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>150</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>21</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
          <LM id="t-ln94210-19-p3s1w23">
           <a>
            <lex.rf>a#a-ln94210-19-p3s1w23</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>roční</t_lemma>
           <functor>THL</functor>
           <tfa>f</tfa>
           <deepord>22</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94210-19-p3s2">
   <atree.rf>a#a-ln94210-19-p3s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94210-19-p3s2w2">
     <a>
      <lex.rf>a#a-ln94210-19-p3s2w2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>vyrábět</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w8288f1</val_frame.rf>
     <children>
      <LM id="t-ln94210-19-p3s2w1">
       <a>
        <lex.rf>a#a-ln94210-19-p3s2w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>společnost</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln94210-19-p3s2w3">
       <a>
        <lex.rf>a#a-ln94210-19-p3s2w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>model</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94210-19-p3s2w4">
         <a>
          <lex.rf>a#a-ln94210-19-p3s2w4</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Charade</t_lemma>
         <functor>ID</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>nr</number>
         </gram>
        </LM>
        <LM id="t-ln94210-19-p3s2w6">
         <a>
          <lex.rf>a#a-ln94210-19-p3s2w6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>automobilka</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94210-19-p3s2w7">
           <a>
            <lex.rf>a#a-ln94210-19-p3s2w7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Daihatsu</t_lemma>
           <functor>ID</functor>
           <tfa>f</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>nr</number>
           </gram>
          </LM>
          <LM id="t-ln94210-19-p3s2w5">
           <a>
            <lex.rf>a#a-ln94210-19-p3s2w5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>japonský</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94210-19-p3s2w13">
         <a>
          <lex.rf>a#a-ln94210-19-p3s2w13</lex.rf>
          <aux.rf>a#a-ln94210-19-p3s2w10</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>používat</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>11</deepord>
         <gram>
          <sempos>v</sempos>
          <verbmod>ind</verbmod>
          <deontmod>decl</deontmod>
          <tense>sim</tense>
          <aspect>proc</aspect>
          <resultative>res0</resultative>
          <dispmod>disp0</dispmod>
          <iterativeness>it0</iterativeness>
         </gram>
         <val_frame.rf>v#v-w4135f1</val_frame.rf>
         <children>
          <LM id="t-ln94210-19-p3s2w9">
           <a>
            <lex.rf>a#a-ln94210-19-p3s2w9</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>který</t_lemma>
           <functor>PAT</functor>
           <tfa>t</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>n.pron.indef</sempos>
            <gender>inher</gender>
            <number>inher</number>
            <indeftype>relat</indeftype>
            <person>inher</person>
           </gram>
           <coref_gram.rf>t-ln94210-19-p3s2w3</coref_gram.rf>
          </LM>
          <LM id="t-ln94210-19-p3s2a0">
           <nodetype>qcomplex</nodetype>
           <is_generated>1</is_generated>
           <t_lemma>#Gen</t_lemma>
           <functor>ACT</functor>
           <tfa>t</tfa>
           <deepord>9</deepord>
          </LM>
          <LM id="t-ln94210-19-p3s2w12">
           <a>
            <lex.rf>a#a-ln94210-19-p3s2w12</lex.rf>
            <aux.rf>a#a-ln94210-19-p3s2w11</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Čína</t_lemma>
           <functor>LOC</functor>
           <subfunctor>basic</subfunctor>
           <tfa>t</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
          </LM>
          <LM id="t-ln94210-19-p3s2w15">
           <a>
            <lex.rf>a#a-ln94210-19-p3s2w15</lex.rf>
            <aux.rf>a#a-ln94210-19-p3s2w14</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>jiný</t_lemma>
           <functor>RESTR</functor>
           <tfa>f</tfa>
           <deepord>12</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>sg</number>
           </gram>
          </LM>
          <LM id="t-ln94210-19-p3s2w17">
           <a>
            <lex.rf>a#a-ln94210-19-p3s2w17</lex.rf>
            <aux.rf>a#a-ln94210-19-p3s2w16</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>taxi</t_lemma>
           <functor>COMPL</functor>
           <tfa>f</tfa>
           <deepord>13</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>nr</number>
           </gram>
           <compl.rf>t-ln94210-19-p3s2w9</compl.rf>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95045-100-p1s1">
   <atree.rf>a#a-ln95045-100-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95045-100-p1s1w2">
     <a>
      <lex.rf>a#a-ln95045-100-p1s1w2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>kapitola</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>fem</gender>
      <number>sg</number>
     </gram>
     <children>
      <LM id="t-ln95045-100-p1s1w1">
       <a>
        <lex.rf>a#a-ln95045-100-p1s1w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>uzavřený</t_lemma>
       <functor>RSTR</functor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>adj.denot</sempos>
        <degcmp>pos</degcmp>
        <negation>neg0</negation>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95045-100-p2s1">
   <atree.rf>a#a-ln95045-100-p2s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95045-100-p2s1w10">
     <a>
      <lex.rf>a#a-ln95045-100-p2s1w10</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>přibýt</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>7</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>post</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w5092f2</val_frame.rf>
     <children>
      <LM id="t-ln95045-100-p2s1w4">
       <a>
        <lex.rf>a#a-ln95045-100-p2s1w4</lex.rf>
        <aux.rf>a#a-ln95045-100-p2s1w1</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>březen</t_lemma>
       <functor>TSIN</functor>
       <tfa>t</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95045-100-p2s1w6">
         <a>
          <lex.rf>a#a-ln95045-100-p2s1w6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>rok</t_lemma>
         <functor>APP</functor>
         <tfa>t</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95045-100-p2s1w5">
           <a>
            <lex.rf>a#a-ln95045-100-p2s1w5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>tento</t_lemma>
           <functor>RSTR</functor>
           <tfa>t</tfa>
           <deepord>1</deepord>
           <gram>
            <sempos>adj.pron.def.demon</sempos>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln95045-100-p2s1w2">
         <a>
          <lex.rf>a#a-ln95045-100-p2s1w2</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>1</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>ord</numertype>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95045-100-p2s1w9">
       <a>
        <lex.rf>a#a-ln95045-100-p2s1w9</lex.rf>
        <aux.rf>a#a-ln95045-100-p2s1w7</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>burza</t_lemma>
       <functor>DIR3</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95045-100-p2s1w8">
         <a>
          <lex.rf>a#a-ln95045-100-p2s1w8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>pražský</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95045-100-p2s1w12">
       <a>
        <lex.rf>a#a-ln95045-100-p2s1w12</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>emise</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>8</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln95045-100-p2s1w13">
         <a>
          <lex.rf>a#a-ln95045-100-p2s1w13</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>akcie</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>9</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln95045-100-p2s1w16">
           <a>
            <lex.rf>a#a-ln95045-100-p2s1w16</lex.rf>
            <aux.rf>a#a-ln95045-100-p2s1w14</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>vlna</t_lemma>
           <functor>DIR1</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln95045-100-p2s1w18">
             <a>
              <lex.rf>a#a-ln95045-100-p2s1w18</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>privatizace</t_lemma>
             <functor>APP</functor>
             <tfa>f</tfa>
             <deepord>11</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln95045-100-p2s1w17">
               <a>
                <lex.rf>a#a-ln95045-100-p2s1w17</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>kupónový</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>12</deepord>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>pos</degcmp>
                <negation>neg0</negation>
               </gram>
              </LM>
             </children>
            </LM>
            <LM id="t-ln95045-100-p2s1w15">
             <a>
              <lex.rf>a#a-ln95045-100-p2s1w15</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>dva</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>13</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>ord</numertype>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-ln95045-100-p2s1w11">
         <a>
          <lex.rf>a#a-ln95045-100-p2s1w11</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>677</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>14</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95045-100-p2s2">
   <atree.rf>a#a-ln95045-100-p2s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95045-100-p2s2a0">
     <a>
      <lex.rf>a#a-ln95045-100-p2s1w10</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <is_generated>1</is_generated>
     <t_lemma>přibýt</t_lemma>
     <functor>PRED</functor>
     <tfa>t</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>post</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w5092f1</val_frame.rf>
     <children>
      <LM id="t-ln95045-100-p2s2w3">
       <a>
        <lex.rf>a#a-ln95045-100-p2s2w3</lex.rf>
        <aux.rf>
         <LM>a#a-ln95045-100-p2s2w1</LM>
         <LM>a#a-ln95045-100-p2s2w2</LM>
        </aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>#PersPron</t_lemma>
       <functor>ACMP</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.pron.def.pers</sempos>
        <gender>fem</gender>
        <number>pl</number>
        <person>3</person>
        <politeness>basic</politeness>
       </gram>
       <coref_text.rf>t-ln95045-100-p2s1w12</coref_text.rf>
      </LM>
      <LM id="t-ln95045-100-p2s2w4">
       <a>
        <lex.rf>a#a-ln95045-100-p2s2w4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>potom</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>adv.pron.def</sempos>
       </gram>
      </LM>
      <LM id="t-ln95045-100-p2s2w6">
       <a>
        <lex.rf>a#a-ln95045-100-p2s2w6</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>emise</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>4</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln95045-100-p2s2w14">
         <a>
          <lex.rf>a#a-ln95045-100-p2s2w14</lex.rf>
         </a>
         <nodetype>coap</nodetype>
         <t_lemma>ale</t_lemma>
         <functor>ADVS</functor>
         <deepord>10</deepord>
         <children>
          <LM id="t-ln95045-100-p2s2w12">
           <a>
            <lex.rf>a#a-ln95045-100-p2s2w12</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>obchodovat</t_lemma>
           <functor>RSTR</functor>
           <is_member>1</is_member>
           <tfa>f</tfa>
           <deepord>9</deepord>
           <gram>
            <sempos>v</sempos>
            <verbmod>ind</verbmod>
            <deontmod>decl</deontmod>
            <tense>ant</tense>
            <aspect>proc</aspect>
            <resultative>res0</resultative>
            <dispmod>disp0</dispmod>
            <iterativeness>it0</iterativeness>
           </gram>
           <val_frame.rf>v#v-w2483f1</val_frame.rf>
           <children>
            <LM id="t-ln95045-100-p2s2w9">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w9</lex.rf>
              <aux.rf>a#a-ln95045-100-p2s2w8</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>který</t_lemma>
             <functor>PAT</functor>
             <tfa>t</tfa>
             <deepord>5</deepord>
             <gram>
              <sempos>n.pron.indef</sempos>
              <gender>inher</gender>
              <number>inher</number>
              <indeftype>relat</indeftype>
              <person>inher</person>
             </gram>
             <coref_gram.rf>t-ln95045-100-p2s2w6</coref_gram.rf>
            </LM>
            <LM id="t-ln95045-100-p2s2w10">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w10</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>už</t_lemma>
             <functor>TWHEN</functor>
             <subfunctor>basic</subfunctor>
             <tfa>t</tfa>
             <deepord>6</deepord>
             <gram>
              <sempos>adv.denot.ngrad.nneg</sempos>
             </gram>
            </LM>
            <LM id="t-ln95045-100-p2s2w11">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w11</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>burza</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>7</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
            </LM>
            <LM id="t-ln95045-100-p2s2a1">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#Gen</t_lemma>
             <functor>ADDR</functor>
             <tfa>t</tfa>
             <deepord>8</deepord>
            </LM>
           </children>
          </LM>
          <LM id="t-ln95045-100-p2s2w22">
           <a>
            <lex.rf>a#a-ln95045-100-p2s2w22</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>být</t_lemma>
           <functor>RSTR</functor>
           <is_member>1</is_member>
           <tfa>f</tfa>
           <deepord>17</deepord>
           <gram>
            <sempos>v</sempos>
            <verbmod>ind</verbmod>
            <deontmod>decl</deontmod>
            <tense>ant</tense>
            <aspect>proc</aspect>
            <resultative>res0</resultative>
            <dispmod>disp0</dispmod>
            <iterativeness>it0</iterativeness>
           </gram>
           <val_frame.rf>v#v-w243f2</val_frame.rf>
           <children>
            <LM id="t-ln95045-100-p2s2w16">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w16</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>část</t_lemma>
             <functor>ACT</functor>
             <tfa>c</tfa>
             <deepord>12</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln95045-100-p2s2w15">
               <a>
                <lex.rf>a#a-ln95045-100-p2s2w15</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>který</t_lemma>
               <functor>MAT</functor>
               <tfa>t</tfa>
               <deepord>11</deepord>
               <gram>
                <sempos>n.pron.indef</sempos>
                <gender>inher</gender>
                <number>inher</number>
                <indeftype>relat</indeftype>
                <person>inher</person>
               </gram>
               <coref_gram.rf>t-ln95045-100-p2s2w6</coref_gram.rf>
              </LM>
             </children>
            </LM>
            <LM id="t-ln95045-100-p2s2w17">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w17</lex.rf>
             </a>
             <nodetype>atom</nodetype>
             <t_lemma>až</t_lemma>
             <functor>RHEM</functor>
             <tfa>t</tfa>
             <deepord>13</deepord>
            </LM>
            <LM id="t-ln95045-100-p2s2w18">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w18</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>dosud</t_lemma>
             <functor>TTILL</functor>
             <tfa>c</tfa>
             <deepord>14</deepord>
             <gram>
              <sempos>adv.denot.ngrad.nneg</sempos>
             </gram>
            </LM>
            <LM id="t-ln95045-100-p2s2w21">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w21</lex.rf>
              <aux.rf>a#a-ln95045-100-p2s2w19</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>důvod</t_lemma>
             <functor>CAUS</functor>
             <tfa>t</tfa>
             <deepord>15</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>pl</number>
             </gram>
             <children>
              <LM id="t-ln95045-100-p2s2w20">
               <a>
                <lex.rf>a#a-ln95045-100-p2s2w20</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>různý</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>16</deepord>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>sup</degcmp>
                <negation>neg0</negation>
               </gram>
              </LM>
             </children>
            </LM>
            <LM id="t-ln95045-100-p2s2w24">
             <a>
              <lex.rf>a#a-ln95045-100-p2s2w24</lex.rf>
              <aux.rf>a#a-ln95045-100-p2s2w23</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>vlastnictví</t_lemma>
             <functor>LOC</functor>
             <subfunctor>basic</subfunctor>
             <tfa>f</tfa>
             <deepord>18</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>neut</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln95045-100-p2s2w25">
               <a>
                <lex.rf>a#a-ln95045-100-p2s2w25</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>Fond</t_lemma>
               <functor>APP</functor>
               <tfa>f</tfa>
               <deepord>19</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>inan</gender>
                <number>sg</number>
               </gram>
               <children>
                <LM id="t-ln95045-100-p2s2w27">
                 <a>
                  <lex.rf>a#a-ln95045-100-p2s2w27</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>majetek</t_lemma>
                 <functor>APP</functor>
                 <tfa>f</tfa>
                 <deepord>20</deepord>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>inan</gender>
                  <number>sg</number>
                 </gram>
                 <children>
                  <LM id="t-ln95045-100-p2s2w26">
                   <a>
                    <lex.rf>a#a-ln95045-100-p2s2w26</lex.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>národní</t_lemma>
                   <functor>RSTR</functor>
                   <tfa>f</tfa>
                   <deepord>21</deepord>
                   <gram>
                    <sempos>adj.denot</sempos>
                    <degcmp>pos</degcmp>
                    <negation>neg0</negation>
                   </gram>
                  </LM>
                 </children>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-ln95045-100-p2s2w5">
         <a>
          <lex.rf>a#a-ln95045-100-p2s2w5</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>184</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>22</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95045-100-p2s3">
   <atree.rf>a#a-ln95045-100-p2s3</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95045-100-p2s3w8">
     <a>
      <lex.rf>a#a-ln95045-100-p2s3w8</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>dosáhnout</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>7</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>post</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w714f1</val_frame.rf>
     <children>
      <LM id="t-ln95045-100-p2s3w2">
       <a>
        <lex.rf>a#a-ln95045-100-p2s3w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>počet</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95045-100-p2s3w4">
         <a>
          <lex.rf>a#a-ln95045-100-p2s3w4</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>titul</t_lemma>
         <functor>MAT</functor>
         <tfa>t</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln95045-100-p2s3w3">
           <a>
            <lex.rf>a#a-ln95045-100-p2s3w3</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>obchodovaný</t_lemma>
           <functor>RSTR</functor>
           <tfa>t</tfa>
           <deepord>1</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln95045-100-p2s3w1">
         <a>
          <lex.rf>a#a-ln95045-100-p2s3w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>celkový</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95045-100-p2s3w5">
       <a>
        <lex.rf>a#a-ln95045-100-p2s3w5</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>tak</t_lemma>
       <functor>MANN</functor>
       <tfa>t</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>adv.pron.def</sempos>
       </gram>
      </LM>
      <LM id="t-ln95045-100-p2s3w7">
       <a>
        <lex.rf>a#a-ln95045-100-p2s3w7</lex.rf>
        <aux.rf>a#a-ln95045-100-p2s3w6</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>burza</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln95045-100-p2s3w10">
       <a>
        <lex.rf>a#a-ln95045-100-p2s3w10</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>číslo</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>8</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>neut</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95045-100-p2s3w11">
         <a>
          <lex.rf>a#a-ln95045-100-p2s3w11</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>1738</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>9</deepord>
         <gram>
          <sempos>n.quant.def</sempos>
          <gender>nr</gender>
          <number>nr</number>
          <numertype>basic</numertype>
         </gram>
        </LM>
        <LM id="t-ln95045-100-p2s3w9">
         <a>
          <lex.rf>a#a-ln95045-100-p2s3w9</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>úctyhodný</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>10</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95045-100-p3s1">
   <atree.rf>a#a-ln95045-100-p3s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95045-100-p3s1w5">
     <a>
      <lex.rf>a#a-ln95045-100-p3s1w5</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>očekávat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>4</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w2608f1</val_frame.rf>
     <children>
      <LM id="t-ln95045-100-p3s1w2">
       <a>
        <lex.rf>a#a-ln95045-100-p3s1w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>veřejnost</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>fem</gender>
        <number>sg</number>
        <negation>neg0</negation>
       </gram>
       <children>
        <LM id="t-ln95045-100-p3s1w1">
         <a>
          <lex.rf>a#a-ln95045-100-p3s1w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>investorský</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95045-100-p3s1w3">
       <a>
        <lex.rf>a#a-ln95045-100-p3s1w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>teď</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>adv.pron.def</sempos>
       </gram>
      </LM>
      <LM id="t-ln95045-100-p3s1w4">
       <a>
        <lex.rf>a#a-ln95045-100-p3s1w4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>trpělivý</t_lemma>
       <functor>MANN</functor>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>adj.denot</sempos>
        <degcmp>pos</degcmp>
        <negation>neg1</negation>
       </gram>
      </LM>
      <LM id="t-ln95045-100-p3s1w11">
       <a>
        <lex.rf>a#a-ln95045-100-p3s1w11</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>stanovit</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>7</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>post</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w6480f2</val_frame.rf>
       <children>
        <LM id="t-ln95045-100-p3s1w10">
         <a>
          <lex.rf>a#a-ln95045-100-p3s1w10</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>burza</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln95045-100-p3s1w9">
         <a>
          <lex.rf>a#a-ln95045-100-p3s1w9</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>cena</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>8</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln95045-100-p3s1w7">
           <a>
            <lex.rf>a#a-ln95045-100-p3s1w7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>jaký</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>9</deepord>
           <gram>
            <sempos>adj.pron.indef</sempos>
            <indeftype>inter</indeftype>
           </gram>
          </LM>
          <LM id="t-ln95045-100-p3s1w8">
           <a>
            <lex.rf>a#a-ln95045-100-p3s1w8</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>výchozí</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95045-100-p3s2">
   <atree.rf>a#a-ln95045-100-p3s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95045-100-p3s2w3">
     <a>
      <lex.rf>a#a-ln95045-100-p3s2w3</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>být</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>4</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w243f1</val_frame.rf>
     <children>
      <LM id="t-ln95045-100-p3s2w2">
       <a>
        <lex.rf>a#a-ln95045-100-p3s2w2</lex.rf>
       </a>
       <nodetype>atom</nodetype>
       <t_lemma>samozřejmě</t_lemma>
       <functor>ATT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
      </LM>
      <LM id="t-ln95045-100-p3s2w1">
       <a>
        <lex.rf>a#a-ln95045-100-p3s2w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>hádanka</t_lemma>
       <functor>PAT</functor>
       <tfa>t</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln95045-100-p3s2a0">
       <nodetype>atom</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Neg</t_lemma>
       <functor>RHEM</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
      </LM>
      <LM id="t-ln95045-100-p3s2w4">
       <a>
        <lex.rf>a#a-ln95045-100-p3s2w4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>kurs</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln95045-100-p3s2w5">
         <a>
          <lex.rf>a#a-ln95045-100-p3s2w5</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>akcie</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln95045-100-p3s2w6">
           <a>
            <lex.rf>a#a-ln95045-100-p3s2w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>známý</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
           <children>
            <LM id="t-ln95045-100-p3s2w9">
             <a>
              <lex.rf>a#a-ln95045-100-p3s2w9</lex.rf>
              <aux.rf>a#a-ln95045-100-p3s2w7</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>vlna</t_lemma>
             <functor>DIR1</functor>
             <subfunctor>basic</subfunctor>
             <tfa>f</tfa>
             <deepord>8</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln95045-100-p3s2w8">
               <a>
                <lex.rf>a#a-ln95045-100-p3s2w8</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>jeden</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>9</deepord>
               <gram>
                <sempos>adj.quant.def</sempos>
                <numertype>ord</numertype>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
          <LM id="t-ln95045-100-p3s2w17">
           <a>
            <lex.rf>a#a-ln95045-100-p3s2w17</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>obchodovat</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>15</deepord>
           <gram>
            <sempos>v</sempos>
            <verbmod>ind</verbmod>
            <deontmod>decl</deontmod>
            <tense>sim</tense>
            <aspect>proc</aspect>
            <resultative>res0</resultative>
            <dispmod>disp0</dispmod>
            <iterativeness>it0</iterativeness>
           </gram>
           <val_frame.rf>v#v-w2483f1</val_frame.rf>
           <children>
            <LM id="t-ln95045-100-p3s2w12">
             <a>
              <lex.rf>a#a-ln95045-100-p3s2w12</lex.rf>
              <aux.rf>a#a-ln95045-100-p3s2w11</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>který</t_lemma>
             <functor>PAT</functor>
             <tfa>t</tfa>
             <deepord>10</deepord>
             <gram>
              <sempos>n.pron.indef</sempos>
              <gender>inher</gender>
              <number>inher</number>
              <indeftype>relat</indeftype>
              <person>inher</person>
             </gram>
             <coref_gram.rf>t-ln95045-100-p3s2w5</coref_gram.rf>
            </LM>
            <LM id="t-ln95045-100-p3s2a2">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#Gen</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>11</deepord>
            </LM>
            <LM id="t-ln95045-100-p3s2a1">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#Gen</t_lemma>
             <functor>ADDR</functor>
             <tfa>t</tfa>
             <deepord>12</deepord>
            </LM>
            <LM id="t-ln95045-100-p3s2w14">
             <a>
              <lex.rf>a#a-ln95045-100-p3s2w14</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>už</t_lemma>
             <functor>TWHEN</functor>
             <subfunctor>basic</subfunctor>
             <tfa>t</tfa>
             <deepord>13</deepord>
             <gram>
              <sempos>adv.denot.ngrad.nneg</sempos>
             </gram>
            </LM>
            <LM id="t-ln95045-100-p3s2w16">
             <a>
              <lex.rf>a#a-ln95045-100-p3s2w16</lex.rf>
              <aux.rf>a#a-ln95045-100-p3s2w15</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>burza</t_lemma>
             <functor>LOC</functor>
             <subfunctor>basic</subfunctor>
             <tfa>t</tfa>
             <deepord>14</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94200-161-p1s1">
   <atree.rf>a#a-ln94200-161-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94200-161-p1s1w2">
     <a>
      <lex.rf>a#a-ln94200-161-p1s1w2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>rekord</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>inan</gender>
      <number>pl</number>
     </gram>
     <children>
      <LM id="t-ln94200-161-p1s1w1">
       <a>
        <lex.rf>a#a-ln94200-161-p1s1w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>dva</t_lemma>
       <functor>RSTR</functor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>adj.quant.def</sempos>
        <numertype>basic</numertype>
       </gram>
      </LM>
      <LM id="t-ln94200-161-p1s1w3">
       <a>
        <lex.rf>a#a-ln94200-161-p1s1w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>najednou</t_lemma>
       <functor>MANN</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>adv.denot.ngrad.nneg</sempos>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94200-161-p2s1A">
   <atree.rf>a#a-ln94200-161-p2s1A</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94200-161-p2s1Aw2">
     <a>
      <lex.rf>a#a-ln94200-161-p2s1Aw2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>plavat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>4</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w3471f1</val_frame.rf>
     <children>
      <LM id="t-ln94200-161-p2s1Aw1">
       <a>
        <lex.rf>a#a-ln94200-161-p2s1Aw1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Perkins</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <is_name_of_person>1</is_name_of_person>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln94200-161-p2s1Aw4">
       <a>
        <lex.rf>a#a-ln94200-161-p2s1Aw4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>m</t_lemma>
       <functor>PAT</functor>
       <tfa>t</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>nr</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s1Aw3">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Aw3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>1500</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94200-161-p2s1Aw5">
       <a>
        <lex.rf>a#a-ln94200-161-p2s1Aw5</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>průměr</t_lemma>
       <functor>MANN</functor>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s1Aw6">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Aw6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>58.77</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>8</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s1Aw9">
           <a>
            <lex.rf>a#a-ln94200-161-p2s1Aw9</lex.rf>
            <aux.rf>a#a-ln94200-161-p2s1Aw7</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>m</t_lemma>
           <functor>REG</functor>
           <tfa>t</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>nr</number>
           </gram>
           <children>
            <LM id="t-ln94200-161-p2s1Aw8">
             <a>
              <lex.rf>a#a-ln94200-161-p2s1Aw8</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>100</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>7</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94200-161-p2s1B">
   <atree.rf>a#a-ln94200-161-p2s1B</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94200-161-p2s1Bw1">
     <a>
      <lex.rf>a#a-ln94200-161-p2s1Bw1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>Victoria</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>fem</gender>
      <number>sg</number>
     </gram>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94200-161-p2s1C">
   <atree.rf>a#a-ln94200-161-p2s1C</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94200-161-p2s1Cw12">
     <a>
      <lex.rf>a#a-ln94200-161-p2s1Cw12</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>vytvořit</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>10</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w8510f1</val_frame.rf>
     <children>
      <LM id="t-ln94200-161-p2s1Cw3">
       <a>
        <lex.rf>a#a-ln94200-161-p2s1Cw3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>rekord</t_lemma>
       <functor>PAT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s1Cw2">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Cw2</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>světový</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-ln94200-161-p2s1Cw1">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Cw1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>dva</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94200-161-p2s1Cw6">
       <a>
        <lex.rf>a#a-ln94200-161-p2s1Cw6</lex.rf>
        <aux.rf>a#a-ln94200-161-p2s1Cw4</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>závod</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>c</tfa>
       <deepord>4</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s1Cw5">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Cw5</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>jediný</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94200-161-p2s1Cw8">
       <a>
        <lex.rf>a#a-ln94200-161-p2s1Cw8</lex.rf>
        <aux.rf>a#a-ln94200-161-p2s1Cw7</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>hra</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s1Cw10">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Cw10</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>společenství</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s1Cw11">
           <a>
            <lex.rf>a#a-ln94200-161-p2s1Cw11</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>národ</t_lemma>
           <functor>APP</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>pl</number>
           </gram>
          </LM>
          <LM id="t-ln94200-161-p2s1Cw9">
           <a>
            <lex.rf>a#a-ln94200-161-p2s1Cw9</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>britský</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>9</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94200-161-p2s1Cw15">
       <a>
        <lex.rf>a#a-ln94200-161-p2s1Cw15</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Perkins</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>11</deepord>
       <is_name_of_person>1</is_name_of_person>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s1Cw14">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Cw14</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Kieren</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>12</deepord>
         <is_name_of_person>1</is_name_of_person>
         <gram>
          <sempos>n.denot</sempos>
          <gender>anim</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94200-161-p2s1Cw13">
         <a>
          <lex.rf>a#a-ln94200-161-p2s1Cw13</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Australan</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>13</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>anim</gender>
          <number>sg</number>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94200-161-p2s2">
   <atree.rf>a#a-ln94200-161-p2s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94200-161-p2s2w16">
     <a>
      <lex.rf>a#a-ln94200-161-p2s2w16</lex.rf>
     </a>
     <nodetype>coap</nodetype>
     <t_lemma>#Comma</t_lemma>
     <functor>CONJ</functor>
     <deepord>14</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-ln94200-161-p2s2w2">
       <a>
        <lex.rf>a#a-ln94200-161-p2s2w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>vítěz</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s2w1">
         <a>
          <lex.rf>a#a-ln94200-161-p2s2w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>olympijský</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-ln94200-161-p2s2w5">
         <a>
          <lex.rf>a#a-ln94200-161-p2s2w5</lex.rf>
          <aux.rf>a#a-ln94200-161-p2s2w3</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>maratón</t_lemma>
         <functor>REG</functor>
         <tfa>f</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s2w4">
           <a>
            <lex.rf>a#a-ln94200-161-p2s2w4</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>plavecký</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>4</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94200-161-p2s2w6">
       <a>
        <lex.rf>a#a-ln94200-161-p2s2w6</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>zdolat</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>ant</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w9394f1</val_frame.rf>
       <children>
        <LM id="t-ln94200-161-p2s2w8">
         <a>
          <lex.rf>a#a-ln94200-161-p2s2w8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>m</t_lemma>
         <functor>PAT</functor>
         <tfa>t</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>nr</number>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s2w7">
           <a>
            <lex.rf>a#a-ln94200-161-p2s2w7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>1500</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
          <LM id="t-ln94200-161-p2s2w10">
           <a>
            <lex.rf>a#a-ln94200-161-p2s2w10</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>způsob</t_lemma>
           <functor>MANN</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94200-161-p2s2w9">
             <a>
              <lex.rf>a#a-ln94200-161-p2s2w9</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>volný</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>9</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94200-161-p2s2w12">
         <a>
          <lex.rf>a#a-ln94200-161-p2s2w12</lex.rf>
          <aux.rf>a#a-ln94200-161-p2s2w11</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>čas</t_lemma>
         <functor>TPAR</functor>
         <tfa>f</tfa>
         <deepord>10</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s2w14">
           <a>
            <lex.rf>a#a-ln94200-161-p2s2w14</lex.rf>
           </a>
           <nodetype>coap</nodetype>
           <t_lemma>#Colon</t_lemma>
           <functor>OPER</functor>
           <deepord>12</deepord>
           <children>
            <LM id="t-ln94200-161-p2s2w13">
             <a>
              <lex.rf>a#a-ln94200-161-p2s2w13</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>14</t_lemma>
             <functor>RSTR</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>11</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
            </LM>
            <LM id="t-ln94200-161-p2s2w15">
             <a>
              <lex.rf>a#a-ln94200-161-p2s2w15</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>41.66</t_lemma>
             <functor>RSTR</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>13</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94200-161-p2s2w18">
       <a>
        <lex.rf>a#a-ln94200-161-p2s2w18</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>překonat</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>16</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>ant</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w4840f1</val_frame.rf>
       <children>
        <LM id="t-ln94200-161-p2s2w17">
         <a>
          <lex.rf>a#a-ln94200-161-p2s2w17</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>co</t_lemma>
         <functor>MEANS</functor>
         <tfa>t</tfa>
         <deepord>15</deepord>
         <gram>
          <sempos>n.pron.indef</sempos>
          <gender>inher</gender>
          <number>inher</number>
          <indeftype>relat</indeftype>
          <person>inher</person>
         </gram>
         <coref_gram.rf>t-ln94200-161-p2s2w6</coref_gram.rf>
        </LM>
        <LM id="t-ln94200-161-p2s2w20">
         <a>
          <lex.rf>a#a-ln94200-161-p2s2w20</lex.rf>
          <aux.rf>a#a-ln94200-161-p2s2w19</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>1.82</t_lemma>
         <functor>DIFF</functor>
         <tfa>f</tfa>
         <deepord>17</deepord>
         <gram>
          <sempos>n.quant.def</sempos>
          <gender>nr</gender>
          <number>nr</number>
          <numertype>basic</numertype>
         </gram>
        </LM>
        <LM id="t-ln94200-161-p2s2w23">
         <a>
          <lex.rf>a#a-ln94200-161-p2s2w23</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>výkon</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>18</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s2w22">
           <a>
            <lex.rf>a#a-ln94200-161-p2s2w22</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>dobrý</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>19</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>sup</degcmp>
            <negation>neg0</negation>
           </gram>
           <children>
            <LM id="t-ln94200-161-p2s2w21">
             <a>
              <lex.rf>a#a-ln94200-161-p2s2w21</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>#PersPron</t_lemma>
             <functor>APP</functor>
             <tfa>f</tfa>
             <deepord>20</deepord>
             <gram>
              <sempos>n.pron.def.pers</sempos>
              <gender>inher</gender>
              <number>inher</number>
              <person>inher</person>
              <politeness>inher</politeness>
             </gram>
             <coref_gram.rf>t-ln94200-161-p2s2w2</coref_gram.rf>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94200-161-p2s2w25">
           <a>
            <lex.rf>a#a-ln94200-161-p2s2w25</lex.rf>
            <aux.rf>a#a-ln94200-161-p2s2w24</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Barcelona</t_lemma>
           <functor>DIR1</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>21</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94200-161-p2s2w26">
             <a>
              <lex.rf>a#a-ln94200-161-p2s2w26</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>1992</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>22</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94200-161-p2s3">
   <atree.rf>a#a-ln94200-161-p2s3</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94200-161-p2s3w3">
     <a>
      <lex.rf>a#a-ln94200-161-p2s3w3</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>naměřit</t_lemma>
     <functor>PRED</functor>
     <tfa>t</tfa>
     <deepord>4</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w2100f1</val_frame.rf>
     <children>
      <LM id="t-ln94200-161-p2s3w1">
       <a>
        <lex.rf>a#a-ln94200-161-p2s3w1</lex.rf>
       </a>
       <nodetype>atom</nodetype>
       <t_lemma>současně</t_lemma>
       <functor>PREC</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
      </LM>
      <LM id="t-ln94200-161-p2s3w4">
       <a>
        <lex.rf>a#a-ln94200-161-p2s3w4</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>rozhodčí</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>pl</number>
       </gram>
      </LM>
      <LM id="t-ln94200-161-p2s3w2">
       <a>
        <lex.rf>a#a-ln94200-161-p2s3w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>#PersPron</t_lemma>
       <functor>BEN</functor>
       <subfunctor>nr</subfunctor>
       <tfa>t</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.pron.def.pers</sempos>
        <gender>anim</gender>
        <number>sg</number>
        <person>3</person>
        <politeness>basic</politeness>
       </gram>
       <coref_text.rf>t-ln94200-161-p2s2w21</coref_text.rf>
      </LM>
      <LM id="t-ln94200-161-p2s3w5">
       <a>
        <lex.rf>a#a-ln94200-161-p2s3w5</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>mezičas</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>7</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s3w8">
         <a>
          <lex.rf>a#a-ln94200-161-p2s3w8</lex.rf>
          <aux.rf>a#a-ln94200-161-p2s3w6</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>m</t_lemma>
         <functor>REG</functor>
         <tfa>t</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>nr</number>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s3w7">
           <a>
            <lex.rf>a#a-ln94200-161-p2s3w7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>800</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94200-161-p2s3w12">
         <a>
          <lex.rf>a#a-ln94200-161-p2s3w12</lex.rf>
         </a>
         <nodetype>coap</nodetype>
         <t_lemma>#Comma</t_lemma>
         <functor>APPS</functor>
         <deepord>11</deepord>
         <children>
          <LM id="t-ln94200-161-p2s3w10">
           <a>
            <lex.rf>a#a-ln94200-161-p2s3w10</lex.rf>
           </a>
           <nodetype>coap</nodetype>
           <t_lemma>#Colon</t_lemma>
           <functor>OPER</functor>
           <is_member>1</is_member>
           <deepord>9</deepord>
           <children>
            <LM id="t-ln94200-161-p2s3w9">
             <a>
              <lex.rf>a#a-ln94200-161-p2s3w9</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>7</t_lemma>
             <functor>RSTR</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>8</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
            </LM>
            <LM id="t-ln94200-161-p2s3w11">
             <a>
              <lex.rf>a#a-ln94200-161-p2s3w11</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>46.00</t_lemma>
             <functor>RSTR</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>10</deepord>
             <gram>
              <sempos>adj.quant.def</sempos>
              <numertype>basic</numertype>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94200-161-p2s3w14">
           <a>
            <lex.rf>a#a-ln94200-161-p2s3w14</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>být</t_lemma>
           <functor>RSTR</functor>
           <is_member>1</is_member>
           <tfa>f</tfa>
           <deepord>13</deepord>
           <gram>
            <sempos>v</sempos>
            <verbmod>ind</verbmod>
            <deontmod>decl</deontmod>
            <tense>sim</tense>
            <aspect>proc</aspect>
            <resultative>res0</resultative>
            <dispmod>disp0</dispmod>
            <iterativeness>it0</iterativeness>
           </gram>
           <val_frame.rf>v#v-w243f1</val_frame.rf>
           <children>
            <LM id="t-ln94200-161-p2s3w13">
             <a>
              <lex.rf>a#a-ln94200-161-p2s3w13</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>co</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>12</deepord>
             <gram>
              <sempos>n.pron.indef</sempos>
              <gender>inher</gender>
              <number>inher</number>
              <indeftype>relat</indeftype>
              <person>inher</person>
             </gram>
             <coref_gram.rf>t-ln94200-161-p2s3w10</coref_gram.rf>
            </LM>
            <LM id="t-ln94200-161-p2s3w15">
             <a>
              <lex.rf>a#a-ln94200-161-p2s3w15</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>dobrý</t_lemma>
             <functor>PAT</functor>
             <tfa>f</tfa>
             <deepord>16</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>comp</degcmp>
              <negation>neg0</negation>
             </gram>
             <children>
              <LM id="t-ln94200-161-p2s3w18">
               <a>
                <lex.rf>a#a-ln94200-161-p2s3w18</lex.rf>
                <aux.rf>a#a-ln94200-161-p2s3w16</aux.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>sekunda</t_lemma>
               <functor>DIFF</functor>
               <tfa>t</tfa>
               <deepord>14</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>nr</number>
               </gram>
               <children>
                <LM id="t-ln94200-161-p2s3w17">
                 <a>
                  <lex.rf>a#a-ln94200-161-p2s3w17</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>0.6</t_lemma>
                 <functor>RSTR</functor>
                 <tfa>f</tfa>
                 <deepord>15</deepord>
                 <gram>
                  <sempos>adj.quant.def</sempos>
                  <numertype>basic</numertype>
                 </gram>
                </LM>
               </children>
              </LM>
              <LM id="t-ln94200-161-p2s3a0">
               <a>
                <lex.rf>a#a-ln94200-161-p2s3w14</lex.rf>
                <aux.rf>a#a-ln94200-161-p2s3w19</aux.rf>
               </a>
               <nodetype>complex</nodetype>
               <is_generated>1</is_generated>
               <t_lemma>být</t_lemma>
               <functor>CPR</functor>
               <subfunctor>basic</subfunctor>
               <tfa>t</tfa>
               <deepord>18</deepord>
               <gram>
                <sempos>v</sempos>
                <verbmod>ind</verbmod>
                <deontmod>decl</deontmod>
                <tense>sim</tense>
                <aspect>proc</aspect>
                <resultative>res0</resultative>
                <dispmod>disp0</dispmod>
                <iterativeness>it0</iterativeness>
               </gram>
               <val_frame.rf>v#v-w243f1</val_frame.rf>
               <children>
                <LM id="t-ln94200-161-p2s3a1">
                 <a>
                  <lex.rf>a#a-ln94200-161-p2s3w15</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <is_generated>1</is_generated>
                 <t_lemma>dobrý</t_lemma>
                 <functor>PAT</functor>
                 <tfa>t</tfa>
                 <deepord>17</deepord>
                 <gram>
                  <sempos>adj.denot</sempos>
                  <degcmp>pos</degcmp>
                  <negation>neg0</negation>
                 </gram>
                </LM>
                <LM id="t-ln94200-161-p2s3w22">
                 <a>
                  <lex.rf>a#a-ln94200-161-p2s3w22</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>rekord</t_lemma>
                 <functor>ACT</functor>
                 <tfa>f</tfa>
                 <deepord>20</deepord>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>inan</gender>
                  <number>sg</number>
                 </gram>
                 <children>
                  <LM id="t-ln94200-161-p2s3w20">
                   <a>
                    <lex.rf>a#a-ln94200-161-p2s3w20</lex.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>#PersPron</t_lemma>
                   <functor>ACT</functor>
                   <tfa>t</tfa>
                   <deepord>19</deepord>
                   <gram>
                    <sempos>n.pron.def.pers</sempos>
                    <gender>anim</gender>
                    <number>sg</number>
                    <person>3</person>
                    <politeness>basic</politeness>
                   </gram>
                   <coref_text.rf>t-ln94200-161-p2s3w2</coref_text.rf>
                  </LM>
                  <LM id="t-ln94200-161-p2s3w21">
                   <a>
                    <lex.rf>a#a-ln94200-161-p2s3w21</lex.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>vlastní</t_lemma>
                   <functor>RSTR</functor>
                   <tfa>f</tfa>
                   <deepord>21</deepord>
                   <gram>
                    <sempos>adj.denot</sempos>
                    <degcmp>pos</degcmp>
                    <negation>neg0</negation>
                   </gram>
                   <val_frame.rf>v#v-w7646f1</val_frame.rf>
                  </LM>
                  <LM id="t-ln94200-161-p2s3w24">
                   <a>
                    <lex.rf>a#a-ln94200-161-p2s3w24</lex.rf>
                    <aux.rf>a#a-ln94200-161-p2s3w23</aux.rf>
                   </a>
                   <nodetype>complex</nodetype>
                   <t_lemma>únor</t_lemma>
                   <functor>TSIN</functor>
                   <tfa>f</tfa>
                   <deepord>22</deepord>
                   <gram>
                    <sempos>n.denot</sempos>
                    <gender>inan</gender>
                    <number>sg</number>
                   </gram>
                   <children>
                    <LM id="t-ln94200-161-p2s3w25">
                     <a>
                      <lex.rf>a#a-ln94200-161-p2s3w25</lex.rf>
                     </a>
                     <nodetype>complex</nodetype>
                     <t_lemma>1992</t_lemma>
                     <functor>APP</functor>
                     <tfa>f</tfa>
                     <deepord>23</deepord>
                     <gram>
                      <sempos>n.quant.def</sempos>
                      <gender>nr</gender>
                      <number>nr</number>
                      <numertype>basic</numertype>
                     </gram>
                    </LM>
                   </children>
                  </LM>
                 </children>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94200-161-p2s4">
   <atree.rf>a#a-ln94200-161-p2s4</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94200-161-p2s4w2">
     <a>
      <lex.rf>a#a-ln94200-161-p2s4w2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>dosáhnout</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w714f1</val_frame.rf>
     <children>
      <LM id="t-ln94200-161-p2s4w1">
       <a>
        <lex.rf>a#a-ln94200-161-p2s4w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Perkins</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <is_name_of_person>1</is_name_of_person>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln94200-161-p2s4w3">
       <a>
        <lex.rf>a#a-ln94200-161-p2s4w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>průměr</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94200-161-p2s4w4">
         <a>
          <lex.rf>a#a-ln94200-161-p2s4w4</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>58.77</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>adj.quant.def</sempos>
          <numertype>basic</numertype>
         </gram>
         <children>
          <LM id="t-ln94200-161-p2s4w6">
           <a>
            <lex.rf>a#a-ln94200-161-p2s4w6</lex.rf>
            <aux.rf>a#a-ln94200-161-p2s4w5</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>sto</t_lemma>
           <functor>REG</functor>
           <tfa>f</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.quant.def</sempos>
            <gender>neut</gender>
            <number>sg</number>
            <numertype>basic</numertype>
           </gram>
           <children>
            <LM id="t-ln94200-161-p2s4w7">
             <a>
              <lex.rf>a#a-ln94200-161-p2s4w7</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>metr</t_lemma>
             <functor>MAT</functor>
             <tfa>t</tfa>
             <deepord>5</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>pl</number>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95049-140-p1s1">
   <atree.rf>a#a-ln95049-140-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95049-140-p1s1w1">
     <a>
      <lex.rf>a#a-ln95049-140-p1s1w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>program</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>inan</gender>
      <number>sg</number>
     </gram>
     <children>
      <LM id="t-ln95049-140-p1s1w2">
       <a>
        <lex.rf>a#a-ln95049-140-p1s1w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>odpovědnost</t_lemma>
       <functor>APP</functor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>fem</gender>
        <number>sg</number>
        <negation>neg0</negation>
       </gram>
       <children>
        <LM id="t-ln95049-140-p1s1w4">
         <a>
          <lex.rf>a#a-ln95049-140-p1s1w4</lex.rf>
          <aux.rf>a#a-ln95049-140-p1s1w3</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>chemie</t_lemma>
         <functor>LOC</functor>
         <subfunctor>basic</subfunctor>
         <tfa>f</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95049-140-p2s1A">
   <atree.rf>a#a-ln95049-140-p2s1A</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95049-140-p2s1Aw1">
     <a>
      <lex.rf>a#a-ln95049-140-p2s1Aw1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>Praha</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>fem</gender>
      <number>sg</number>
     </gram>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95049-140-p2s1B">
   <atree.rf>a#a-ln95049-140-p2s1B</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95049-140-p2s1Bw14">
     <a>
      <lex.rf>a#a-ln95049-140-p2s1Bw14</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>rozeslat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>14</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w5616f1</val_frame.rf>
     <children>
      <LM id="t-ln95049-140-p2s1Bw1">
       <a>
        <lex.rf>a#a-ln95049-140-p2s1Bw1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>zásada</t_lemma>
       <functor>PAT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p2s1Bw3">
         <a>
          <lex.rf>a#a-ln95049-140-p2s1Bw3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>program</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95049-140-p2s1Bw8">
           <a>
            <lex.rf>a#a-ln95049-140-p2s1Bw8</lex.rf>
           </a>
           <nodetype>coap</nodetype>
           <t_lemma>#Dash</t_lemma>
           <functor>APPS</functor>
           <deepord>7</deepord>
           <children>
            <LM id="t-ln95049-140-p2s1Bw5">
             <a>
              <lex.rf>a#a-ln95049-140-p2s1Bw5</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>podnikání</t_lemma>
             <functor>ID</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>4</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>neut</gender>
              <number>sg</number>
             </gram>
             <val_frame.rf>v#v-w3600f1</val_frame.rf>
             <children>
              <LM id="t-ln95049-140-p2s1Ba0">
               <nodetype>qcomplex</nodetype>
               <is_generated>1</is_generated>
               <t_lemma>#Gen</t_lemma>
               <functor>ACT</functor>
               <tfa>t</tfa>
               <deepord>3</deepord>
              </LM>
              <LM id="t-ln95049-140-p2s1Bw4">
               <a>
                <lex.rf>a#a-ln95049-140-p2s1Bw4</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>Odpovědný</t_lemma>
               <functor>MANN</functor>
               <tfa>f</tfa>
               <deepord>5</deepord>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>pos</degcmp>
                <negation>neg0</negation>
               </gram>
              </LM>
              <LM id="t-ln95049-140-p2s1Bw7">
               <a>
                <lex.rf>a#a-ln95049-140-p2s1Bw7</lex.rf>
                <aux.rf>a#a-ln95049-140-p2s1Bw6</aux.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>chemie</t_lemma>
               <functor>LOC</functor>
               <subfunctor>basic</subfunctor>
               <tfa>f</tfa>
               <deepord>6</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>sg</number>
               </gram>
              </LM>
             </children>
            </LM>
            <LM id="t-ln95049-140-p2s1Bw11">
             <a>
              <lex.rf>a#a-ln95049-140-p2s1Bw11</lex.rf>
             </a>
             <nodetype>coap</nodetype>
             <t_lemma>#Bracket</t_lemma>
             <functor>APPS</functor>
             <is_member>1</is_member>
             <deepord>11</deepord>
             <children>
              <LM id="t-ln95049-140-p2s1Ba1">
               <nodetype>list</nodetype>
               <is_generated>1</is_generated>
               <t_lemma>#Forn</t_lemma>
               <functor>ID</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>8</deepord>
               <children>
                <LM id="t-ln95049-140-p2s1Bw9">
                 <a>
                  <lex.rf>a#a-ln95049-140-p2s1Bw9</lex.rf>
                 </a>
                 <nodetype>fphr</nodetype>
                 <t_lemma>Responsible</t_lemma>
                 <functor>FPHR</functor>
                 <deepord>9</deepord>
                </LM>
                <LM id="t-ln95049-140-p2s1Bw10">
                 <a>
                  <lex.rf>a#a-ln95049-140-p2s1Bw10</lex.rf>
                 </a>
                 <nodetype>fphr</nodetype>
                 <t_lemma>Care</t_lemma>
                 <functor>FPHR</functor>
                 <deepord>10</deepord>
                </LM>
               </children>
              </LM>
              <LM id="t-ln95049-140-p2s1Bw12">
               <a>
                <lex.rf>a#a-ln95049-140-p2s1Bw12</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>RC</t_lemma>
               <functor>ID</functor>
               <is_member>1</is_member>
               <tfa>f</tfa>
               <deepord>12</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>nr</number>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
          <LM id="t-ln95049-140-p2s1Bw2">
           <a>
            <lex.rf>a#a-ln95049-140-p2s1Bw2</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>mezinárodní</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>13</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p2s1Bw15">
       <a>
        <lex.rf>a#a-ln95049-140-p2s1Bw15</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Svaz</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>15</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p2s1Bw17">
         <a>
          <lex.rf>a#a-ln95049-140-p2s1Bw17</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>průmysl</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>16</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95049-140-p2s1Bw16">
           <a>
            <lex.rf>a#a-ln95049-140-p2s1Bw16</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>chemický</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>17</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p2s1Bw19">
       <a>
        <lex.rf>a#a-ln95049-140-p2s1Bw19</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>společnost</t_lemma>
       <functor>ADDR</functor>
       <tfa>f</tfa>
       <deepord>18</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p2s1Bw18">
         <a>
          <lex.rf>a#a-ln95049-140-p2s1Bw18</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>jednotlivý</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>19</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-ln95049-140-p2s1Bw25">
         <a>
          <lex.rf>a#a-ln95049-140-p2s1Bw25</lex.rf>
          <aux.rf>a#a-ln95049-140-p2s1Bw22</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>přihlásit_se</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>22</deepord>
         <gram>
          <sempos>v</sempos>
          <verbmod>ind</verbmod>
          <deontmod>decl</deontmod>
          <tense>ant</tense>
          <aspect>cpl</aspect>
          <resultative>res0</resultative>
          <dispmod>disp0</dispmod>
          <iterativeness>it0</iterativeness>
         </gram>
         <val_frame.rf>v#v-w5120f1</val_frame.rf>
         <children>
          <LM id="t-ln95049-140-p2s1Bw21">
           <a>
            <lex.rf>a#a-ln95049-140-p2s1Bw21</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>který</t_lemma>
           <functor>ACT</functor>
           <tfa>t</tfa>
           <deepord>20</deepord>
           <gram>
            <sempos>n.pron.indef</sempos>
            <gender>inher</gender>
            <number>inher</number>
            <indeftype>relat</indeftype>
            <person>inher</person>
           </gram>
           <coref_gram.rf>t-ln95049-140-p2s1Bw19</coref_gram.rf>
          </LM>
          <LM id="t-ln95049-140-p2s1Bw24">
           <a>
            <lex.rf>a#a-ln95049-140-p2s1Bw24</lex.rf>
            <aux.rf>a#a-ln95049-140-p2s1Bw23</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>RC</t_lemma>
           <functor>PAT</functor>
           <tfa>t</tfa>
           <deepord>21</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>nr</number>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p2s1Bw28">
       <a>
        <lex.rf>a#a-ln95049-140-p2s1Bw28</lex.rf>
        <aux.rf>a#a-ln95049-140-p2s1Bw27</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>prosinec</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>23</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p2s1Bw30">
         <a>
          <lex.rf>a#a-ln95049-140-p2s1Bw30</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>rok</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>24</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95049-140-p2s1Bw29">
           <a>
            <lex.rf>a#a-ln95049-140-p2s1Bw29</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>minulý</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>25</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95049-140-p2s2">
   <atree.rf>a#a-ln95049-140-p2s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95049-140-p2s2w9">
     <a>
      <lex.rf>a#a-ln95049-140-p2s2w9</lex.rf>
      <aux.rf>
       <LM>a#a-ln95049-140-p2s2w7</LM>
       <LM>a#a-ln95049-140-p2s2w8</LM>
       <LM>a#a-ln95049-140-p2s2w6</LM>
      </aux.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>zpracovat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>7</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>cdn</verbmod>
      <deontmod>hrt</deontmod>
      <tense>sim</tense>
      <aspect>cpl</aspect>
      <resultative>res1</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w9810f1</val_frame.rf>
     <children>
      <LM id="t-ln95049-140-p2s2w4">
       <a>
        <lex.rf>a#a-ln95049-140-p2s2w4</lex.rf>
       </a>
       <nodetype>coap</nodetype>
       <t_lemma>až</t_lemma>
       <functor>OPER</functor>
       <deepord>2</deepord>
       <children>
        <LM id="t-ln95049-140-p2s2w3">
         <a>
          <lex.rf>a#a-ln95049-140-p2s2w3</lex.rf>
          <aux.rf>a#a-ln95049-140-p2s2w1</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>březen</t_lemma>
         <functor>TPAR</functor>
         <is_member>1</is_member>
         <tfa>t</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln95049-140-p2s2w5">
         <a>
          <lex.rf>a#a-ln95049-140-p2s2w5</lex.rf>
          <aux.rf>a#a-ln95049-140-p2s2w1</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>duben</t_lemma>
         <functor>TPAR</functor>
         <is_member>1</is_member>
         <tfa>t</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln95049-140-p2s2w2">
         <a>
          <lex.rf>a#a-ln95049-140-p2s2w2</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>letošní</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p2s2a0">
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Gen</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>5</deepord>
      </LM>
      <LM id="t-ln95049-140-p2s2a1">
       <nodetype>complex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#PersPron</t_lemma>
       <functor>PAT</functor>
       <tfa>t</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>n.pron.def.pers</sempos>
        <gender>inan</gender>
        <number>sg</number>
        <person>3</person>
        <politeness>basic</politeness>
       </gram>
       <coref_text.rf>t-ln95049-140-p2s1Bw3</coref_text.rf>
      </LM>
      <LM id="t-ln95049-140-p2s2w12">
       <a>
        <lex.rf>a#a-ln95049-140-p2s2w12</lex.rf>
        <aux.rf>a#a-ln95049-140-p2s2w10</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>program</t_lemma>
       <functor>COMPL</functor>
       <tfa>f</tfa>
       <deepord>8</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <compl.rf>t-ln95049-140-p2s2a1</compl.rf>
       <children>
        <LM id="t-ln95049-140-p2s2w11">
         <a>
          <lex.rf>a#a-ln95049-140-p2s2w11</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>svazový</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>9</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95049-140-p3s1">
   <atree.rf>a#a-ln95049-140-p3s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95049-140-p3s1w15">
     <a>
      <lex.rf>a#a-ln95049-140-p3s1w15</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>vyhlásit</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>14</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w7884f1</val_frame.rf>
     <children>
      <LM id="t-ln95049-140-p3s1w2">
       <a>
        <lex.rf>a#a-ln95049-140-p3s1w2</lex.rf>
        <aux.rf>a#a-ln95049-140-p3s1w1</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>sdělit</t_lemma>
       <functor>PAR</functor>
       <is_parenthesis>1</is_parenthesis>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <sentmod>enunc</sentmod>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>ant</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w5918f1</val_frame.rf>
       <children>
        <LM id="t-ln95049-140-p3s1a1">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ADDR</functor>
         <is_parenthesis>1</is_parenthesis>
         <tfa>t</tfa>
         <deepord>1</deepord>
        </LM>
        <LM id="t-ln95049-140-p3s1a0">
         <nodetype>complex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#PersPron</t_lemma>
         <functor>EFF</functor>
         <is_parenthesis>1</is_parenthesis>
         <tfa>t</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.pron.def.pers</sempos>
          <gender>neut</gender>
          <number>sg</number>
          <person>3</person>
          <politeness>basic</politeness>
         </gram>
         <coref_text.rf>t-ln95049-140-p3s1w15</coref_text.rf>
        </LM>
        <LM id="t-ln95049-140-p3s1w10">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w10</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Zíka</t_lemma>
         <functor>ACT</functor>
         <is_parenthesis>1</is_parenthesis>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <is_name_of_person>1</is_name_of_person>
         <gram>
          <sempos>n.denot</sempos>
          <gender>anim</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95049-140-p3s1w9">
           <a>
            <lex.rf>a#a-ln95049-140-p3s1w9</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Ivan</t_lemma>
           <functor>RSTR</functor>
           <is_parenthesis>1</is_parenthesis>
           <tfa>f</tfa>
           <deepord>5</deepord>
           <is_name_of_person>1</is_name_of_person>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
          </LM>
          <LM id="t-ln95049-140-p3s1w3">
           <a>
            <lex.rf>a#a-ln95049-140-p3s1w3</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>spolupracovník</t_lemma>
           <functor>RSTR</functor>
           <is_parenthesis>1</is_parenthesis>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln95049-140-p3s1w4">
             <a>
              <lex.rf>a#a-ln95049-140-p3s1w4</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>svaz</t_lemma>
             <functor>APP</functor>
             <is_parenthesis>1</is_parenthesis>
             <tfa>t</tfa>
             <deepord>6</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
            </LM>
            <LM id="t-ln95049-140-p3s1w6">
             <a>
              <lex.rf>a#a-ln95049-140-p3s1w6</lex.rf>
              <aux.rf>a#a-ln95049-140-p3s1w5</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>otázka</t_lemma>
             <functor>BEN</functor>
             <subfunctor>basic</subfunctor>
             <is_parenthesis>1</is_parenthesis>
             <tfa>f</tfa>
             <deepord>8</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>pl</number>
             </gram>
             <children>
              <LM id="t-ln95049-140-p3s1w7">
               <a>
                <lex.rf>a#a-ln95049-140-p3s1w7</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>ochrana</t_lemma>
               <functor>PAT</functor>
               <is_parenthesis>1</is_parenthesis>
               <tfa>f</tfa>
               <deepord>9</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>sg</number>
               </gram>
               <children>
                <LM id="t-ln95049-140-p3s1w8">
                 <a>
                  <lex.rf>a#a-ln95049-140-p3s1w8</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>prostředí</t_lemma>
                 <functor>PAT</functor>
                 <is_parenthesis>1</is_parenthesis>
                 <tfa>f</tfa>
                 <deepord>10</deepord>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>neut</gender>
                  <number>sg</number>
                 </gram>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p3s1w13">
       <a>
        <lex.rf>a#a-ln95049-140-p3s1w13</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>program</t_lemma>
       <functor>PAT</functor>
       <tfa>t</tfa>
       <deepord>12</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p3s1w14">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w14</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>RC</t_lemma>
         <functor>ID</functor>
         <tfa>t</tfa>
         <deepord>11</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>nr</number>
         </gram>
        </LM>
        <LM id="t-ln95049-140-p3s1w12">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w12</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>dobrovolný</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>13</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p3s1w17">
       <a>
        <lex.rf>a#a-ln95049-140-p3s1w17</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>organizace</t_lemma>
       <functor>ACT</functor>
       <tfa>f</tfa>
       <deepord>15</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p3s1w20">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w20</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>CEFIC</t_lemma>
         <functor>ID</functor>
         <tfa>f</tfa>
         <deepord>16</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>nr</number>
         </gram>
        </LM>
        <LM id="t-ln95049-140-p3s1w18">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w18</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>svaz</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>17</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>pl</number>
         </gram>
         <children>
          <LM id="t-ln95049-140-p3s1w19">
           <a>
            <lex.rf>a#a-ln95049-140-p3s1w19</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>chemie</t_lemma>
           <functor>APP</functor>
           <tfa>f</tfa>
           <deepord>18</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln95049-140-p3s1w16">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w16</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>mezinárodní</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>19</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p3s1w22">
       <a>
        <lex.rf>a#a-ln95049-140-p3s1w22</lex.rf>
        <aux.rf>a#a-ln95049-140-p3s1w21</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>rok</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>20</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p3s1w23">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w23</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>1986</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>21</deepord>
         <gram>
          <sempos>n.quant.def</sempos>
          <gender>nr</gender>
          <number>nr</number>
          <numertype>basic</numertype>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95049-140-p3s1w26">
       <a>
        <lex.rf>a#a-ln95049-140-p3s1w26</lex.rf>
        <aux.rf>
         <LM>a#a-ln95049-140-p3s1w24</LM>
         <LM>a#a-ln95049-140-p3s1w25</LM>
        </aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>zkušenost</t_lemma>
       <functor>CRIT</functor>
       <tfa>f</tfa>
       <deepord>22</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>fem</gender>
        <number>pl</number>
        <negation>neg0</negation>
       </gram>
       <children>
        <LM id="t-ln95049-140-p3s1w27">
         <a>
          <lex.rf>a#a-ln95049-140-p3s1w27</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>získaný</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>23</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
         <children>
          <LM id="t-ln95049-140-p3s1w29">
           <a>
            <lex.rf>a#a-ln95049-140-p3s1w29</lex.rf>
            <aux.rf>a#a-ln95049-140-p3s1w28</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>rok</t_lemma>
           <functor>TSIN</functor>
           <tfa>f</tfa>
           <deepord>24</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln95049-140-p3s1w30">
             <a>
              <lex.rf>a#a-ln95049-140-p3s1w30</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>1984</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>25</deepord>
             <gram>
              <sempos>n.quant.def</sempos>
              <gender>nr</gender>
              <number>nr</number>
              <numertype>basic</numertype>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln95049-140-p3s1w32">
           <a>
            <lex.rf>a#a-ln95049-140-p3s1w32</lex.rf>
            <aux.rf>a#a-ln95049-140-p3s1w31</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Kanada</t_lemma>
           <functor>LOC</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>26</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95049-140-p3s2">
   <atree.rf>a#a-ln95049-140-p3s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95049-140-p3s2w2">
     <a>
      <lex.rf>a#a-ln95049-140-p3s2w2</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>navazovat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>2</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w2294f1</val_frame.rf>
     <children>
      <LM id="t-ln95049-140-p3s2w1">
       <a>
        <lex.rf>a#a-ln95049-140-p3s2w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>program</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln95049-140-p3s2w4">
       <a>
        <lex.rf>a#a-ln95049-140-p3s2w4</lex.rf>
        <aux.rf>a#a-ln95049-140-p3s2w3</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>standardizace</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95049-140-p3s2w5">
         <a>
          <lex.rf>a#a-ln95049-140-p3s2w5</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>ISO</t_lemma>
         <functor>ID</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>nr</number>
         </gram>
         <children>
          <LM id="t-ln95049-140-p3s2w6">
           <a>
            <lex.rf>a#a-ln95049-140-p3s2w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>9000</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>5</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>basic</numertype>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95046-082-p1s1">
   <atree.rf>a#a-ln95046-082-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95046-082-p1s1w3">
     <a>
      <lex.rf>a#a-ln95046-082-p1s1w3</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>vypracovat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>3</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>ant</tense>
      <aspect>cpl</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w8236f1</val_frame.rf>
     <children>
      <LM id="t-ln95046-082-p1s1w2">
       <a>
        <lex.rf>a#a-ln95046-082-p1s1w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>demokrat</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>anim</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p1s1w1">
         <a>
          <lex.rf>a#a-ln95046-082-p1s1w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>sociální</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95046-082-p1s1w5">
       <a>
        <lex.rf>a#a-ln95046-082-p1s1w5</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>verze</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>4</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p1s1w4">
         <a>
          <lex.rf>a#a-ln95046-082-p1s1w4</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>základní</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-ln95046-082-p1s1w8">
         <a>
          <lex.rf>a#a-ln95046-082-p1s1w8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>program</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p1s1w6">
           <a>
            <lex.rf>a#a-ln95046-082-p1s1w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>#PersPron</t_lemma>
           <functor>APP</functor>
           <tfa>t</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.pron.def.pers</sempos>
            <gender>inher</gender>
            <number>inher</number>
            <person>inher</person>
            <politeness>inher</politeness>
           </gram>
           <coref_gram.rf>t-ln95046-082-p1s1w2</coref_gram.rf>
          </LM>
          <LM id="t-ln95046-082-p1s1w7">
           <a>
            <lex.rf>a#a-ln95046-082-p1s1w7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>volební</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95046-082-p2s1A">
   <atree.rf>a#a-ln95046-082-p2s1A</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95046-082-p2s1Aw1">
     <a>
      <lex.rf>a#a-ln95046-082-p2s1Aw1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>Praha</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>fem</gender>
      <number>sg</number>
     </gram>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95046-082-p2s1B">
   <atree.rf>a#a-ln95046-082-p2s1B</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95046-082-p2s1Bw11">
     <a>
      <lex.rf>a#a-ln95046-082-p2s1Bw11</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>být</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>10</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>post</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w243f1</val_frame.rf>
     <children>
      <LM id="t-ln95046-082-p2s1Bw7">
       <a>
        <lex.rf>a#a-ln95046-082-p2s1Bw7</lex.rf>
       </a>
       <nodetype>coap</nodetype>
       <t_lemma>a</t_lemma>
       <functor>CONJ</functor>
       <deepord>6</deepord>
       <children>
        <LM id="t-ln95046-082-p2s1Bw1">
         <a>
          <lex.rf>a#a-ln95046-082-p2s1Bw1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>ochrana</t_lemma>
         <functor>ACT</functor>
         <is_member>1</is_member>
         <tfa>c</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p2s1Bw3">
           <a>
            <lex.rf>a#a-ln95046-082-p2s1Bw3</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>prostředí</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>2</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>neut</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln95046-082-p2s1Bw2">
             <a>
              <lex.rf>a#a-ln95046-082-p2s1Bw2</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>životní</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>3</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-ln95046-082-p2s1Bw6">
         <a>
          <lex.rf>a#a-ln95046-082-p2s1Bw6</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>rozvoj</t_lemma>
         <functor>ACT</functor>
         <is_member>1</is_member>
         <tfa>c</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p2s1Bw5">
           <a>
            <lex.rf>a#a-ln95046-082-p2s1Bw5</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>hospodářský</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>5</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln95046-082-p2s1Bw8">
         <a>
          <lex.rf>a#a-ln95046-082-p2s1Bw8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>rozvoj</t_lemma>
         <functor>ACT</functor>
         <is_member>1</is_member>
         <tfa>c</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p2s1Bw10">
           <a>
            <lex.rf>a#a-ln95046-082-p2s1Bw10</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>sféra</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln95046-082-p2s1Bw9">
             <a>
              <lex.rf>a#a-ln95046-082-p2s1Bw9</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>sociální</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>9</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95046-082-p2s1Bw13">
       <a>
        <lex.rf>a#a-ln95046-082-p2s1Bw13</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>pilíř</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>11</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>pl</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p2s1Bw12">
         <a>
          <lex.rf>a#a-ln95046-082-p2s1Bw12</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>hlavní</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>12</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-ln95046-082-p2s1Bw14">
         <a>
          <lex.rf>a#a-ln95046-082-p2s1Bw14</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>politika</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>16</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p2s1Bw17">
           <a>
            <lex.rf>a#a-ln95046-082-p2s1Bw17</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>demokracie</t_lemma>
           <functor>APP</functor>
           <tfa>t</tfa>
           <deepord>15</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln95046-082-p2s1Bw15">
             <a>
              <lex.rf>a#a-ln95046-082-p2s1Bw15</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>český</t_lemma>
             <functor>RSTR</functor>
             <tfa>t</tfa>
             <deepord>13</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
            <LM id="t-ln95046-082-p2s1Bw16">
             <a>
              <lex.rf>a#a-ln95046-082-p2s1Bw16</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>sociální</t_lemma>
             <functor>RSTR</functor>
             <tfa>t</tfa>
             <deepord>14</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95046-082-p2s2">
   <atree.rf>a#a-ln95046-082-p2s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95046-082-p2s2w1">
     <a>
      <lex.rf>a#a-ln95046-082-p2s2w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>uvádět</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>3</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w7420f3</val_frame.rf>
     <children>
      <LM id="t-ln95046-082-p2s2w3">
       <a>
        <lex.rf>a#a-ln95046-082-p2s2w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>ten</t_lemma>
       <functor>PAT</functor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.pron.def.demon</sempos>
        <gender>neut</gender>
        <number>sg</number>
       </gram>
       <coref_text.rf>t-ln95046-082-p2s1Bw11</coref_text.rf>
      </LM>
      <LM id="t-ln95046-082-p2s2a0">
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Gen</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>2</deepord>
      </LM>
      <LM id="t-ln95046-082-p2s2w5">
       <a>
        <lex.rf>a#a-ln95046-082-p2s2w5</lex.rf>
        <aux.rf>a#a-ln95046-082-p2s2w4</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>návrh</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>4</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p2s2w7">
         <a>
          <lex.rf>a#a-ln95046-082-p2s2w7</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>program</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p2s2w6">
           <a>
            <lex.rf>a#a-ln95046-082-p2s2w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>volební</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
          <LM id="t-ln95046-082-p2s2w9">
           <a>
            <lex.rf>a#a-ln95046-082-p2s2w9</lex.rf>
            <aux.rf>a#a-ln95046-082-p2s2w8</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>název</t_lemma>
           <functor>ACMP</functor>
           <subfunctor>basic</subfunctor>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln95046-082-p2s2w11">
             <a>
              <lex.rf>a#a-ln95046-082-p2s2w11</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>Společnost</t_lemma>
             <functor>ID</functor>
             <tfa>f</tfa>
             <deepord>8</deepord>
             <quot>
              <LM>
               <type>title</type>
               <set_id>ln95046-082-p2s2w11</set_id>
              </LM>
             </quot>
             <gram>
              <sempos>n.denot.neg</sempos>
              <gender>fem</gender>
              <number>sg</number>
              <negation>neg0</negation>
             </gram>
             <children>
              <LM id="t-ln95046-082-p2s2w15">
               <a>
                <lex.rf>a#a-ln95046-082-p2s2w15</lex.rf>
               </a>
               <nodetype>coap</nodetype>
               <t_lemma>a</t_lemma>
               <functor>CONJ</functor>
               <deepord>11</deepord>
               <quot>
                <LM>
                 <type>title</type>
                 <set_id>ln95046-082-p2s2w11</set_id>
                </LM>
               </quot>
               <children>
                <LM id="t-ln95046-082-p2s2w12">
                 <a>
                  <lex.rf>a#a-ln95046-082-p2s2w12</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>vzdělání</t_lemma>
                 <functor>APP</functor>
                 <is_member>1</is_member>
                 <tfa>f</tfa>
                 <deepord>9</deepord>
                 <quot>
                  <LM>
                   <type>title</type>
                   <set_id>ln95046-082-p2s2w11</set_id>
                  </LM>
                 </quot>
                 <gram>
                  <sempos>n.denot.neg</sempos>
                  <gender>neut</gender>
                  <number>sg</number>
                  <negation>neg0</negation>
                 </gram>
                 <val_frame.rf>v#v-w8644f1</val_frame.rf>
                </LM>
                <LM id="t-ln95046-082-p2s2w14">
                 <a>
                  <lex.rf>a#a-ln95046-082-p2s2w14</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>spoluúčast</t_lemma>
                 <functor>APP</functor>
                 <is_member>1</is_member>
                 <tfa>f</tfa>
                 <deepord>10</deepord>
                 <quot>
                  <LM>
                   <type>title</type>
                   <set_id>ln95046-082-p2s2w11</set_id>
                  </LM>
                 </quot>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>fem</gender>
                  <number>sg</number>
                 </gram>
                </LM>
                <LM id="t-ln95046-082-p2s2w16">
                 <a>
                  <lex.rf>a#a-ln95046-082-p2s2w16</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>solidarita</t_lemma>
                 <functor>APP</functor>
                 <is_member>1</is_member>
                 <tfa>f</tfa>
                 <deepord>12</deepord>
                 <quot>
                  <LM>
                   <type>title</type>
                   <set_id>ln95046-082-p2s2w11</set_id>
                  </LM>
                 </quot>
                 <gram>
                  <sempos>n.denot</sempos>
                  <gender>fem</gender>
                  <number>sg</number>
                 </gram>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
          <LM id="t-ln95046-082-p2s2w21">
           <a>
            <lex.rf>a#a-ln95046-082-p2s2w21</lex.rf>
            <aux.rf>a#a-ln95046-082-p2s2w20</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>schvalovat</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>14</deepord>
           <gram>
            <sempos>v</sempos>
            <verbmod>ind</verbmod>
            <deontmod>hrt</deontmod>
            <tense>sim</tense>
            <aspect>proc</aspect>
            <resultative>res0</resultative>
            <dispmod>disp0</dispmod>
            <iterativeness>it0</iterativeness>
           </gram>
           <val_frame.rf>v#v-w6057f1</val_frame.rf>
           <children>
            <LM id="t-ln95046-082-p2s2w19">
             <a>
              <lex.rf>a#a-ln95046-082-p2s2w19</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>který</t_lemma>
             <functor>PAT</functor>
             <tfa>t</tfa>
             <deepord>13</deepord>
             <gram>
              <sempos>n.pron.indef</sempos>
              <gender>inher</gender>
              <number>inher</number>
              <indeftype>relat</indeftype>
              <person>inher</person>
             </gram>
             <coref_gram.rf>t-ln95046-082-p2s2w7</coref_gram.rf>
            </LM>
            <LM id="t-ln95046-082-p2s2w23">
             <a>
              <lex.rf>a#a-ln95046-082-p2s2w23</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>sjezd</t_lemma>
             <functor>ACT</functor>
             <tfa>f</tfa>
             <deepord>16</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln95046-082-p2s2w24">
               <a>
                <lex.rf>a#a-ln95046-082-p2s2w24</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>ČSSD</t_lemma>
               <functor>APP</functor>
               <tfa>t</tfa>
               <deepord>15</deepord>
               <gram>
                <sempos>n.denot</sempos>
                <gender>fem</gender>
                <number>nr</number>
               </gram>
              </LM>
              <LM id="t-ln95046-082-p2s2w22">
               <a>
                <lex.rf>a#a-ln95046-082-p2s2w22</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>dubnový</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>17</deepord>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>pos</degcmp>
                <negation>neg0</negation>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95046-082-p3s1">
   <atree.rf>a#a-ln95046-082-p3s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95046-082-p3s1w9">
     <a>
      <lex.rf>a#a-ln95046-082-p3s1w9</lex.rf>
      <aux.rf>a#a-ln95046-082-p3s1w5</aux.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>napomáhat</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>7</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>deb</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w2148f2</val_frame.rf>
     <children>
      <LM id="t-ln95046-082-p3s1w1">
       <a>
        <lex.rf>a#a-ln95046-082-p3s1w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>politika</t_lemma>
       <functor>ACT</functor>
       <tfa>c</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p3s1w4">
         <a>
          <lex.rf>a#a-ln95046-082-p3s1w4</lex.rf>
          <aux.rf>a#a-ln95046-082-p3s1w2</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>oblast</t_lemma>
         <functor>LOC</functor>
         <subfunctor>basic</subfunctor>
         <tfa>f</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p3s1w3">
           <a>
            <lex.rf>a#a-ln95046-082-p3s1w3</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>sociální</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>3</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95046-082-p3s1w7">
       <a>
        <lex.rf>a#a-ln95046-082-p3s1w7</lex.rf>
        <aux.rf>a#a-ln95046-082-p3s1w6</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>návrh</t_lemma>
       <functor>CRIT</functor>
       <tfa>t</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p3s1w8">
         <a>
          <lex.rf>a#a-ln95046-082-p3s1w8</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>ČSSD</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>nr</number>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95046-082-p3s1a1">
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Gen</t_lemma>
       <functor>ADDR</functor>
       <tfa>t</tfa>
       <deepord>6</deepord>
      </LM>
      <LM id="t-ln95046-082-p3s1w11">
       <a>
        <lex.rf>a#a-ln95046-082-p3s1w11</lex.rf>
        <aux.rf>a#a-ln95046-082-p3s1w10</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>získání</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>9</deepord>
       <gram>
        <sempos>n.denot.neg</sempos>
        <gender>neut</gender>
        <number>sg</number>
        <negation>neg0</negation>
       </gram>
       <val_frame.rf>v#v-w9499f1</val_frame.rf>
       <children>
        <LM id="t-ln95046-082-p3s1a0">
         <nodetype>qcomplex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#Gen</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>8</deepord>
        </LM>
        <LM id="t-ln95046-082-p3s1w13">
         <a>
          <lex.rf>a#a-ln95046-082-p3s1w13</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>odměna</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>10</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p3s1w12">
           <a>
            <lex.rf>a#a-ln95046-082-p3s1w12</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>spravedlivý</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>11</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
          <LM id="t-ln95046-082-p3s1w15">
           <a>
            <lex.rf>a#a-ln95046-082-p3s1w15</lex.rf>
            <aux.rf>a#a-ln95046-082-p3s1w14</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>práce</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>12</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95046-082-p3s2">
   <atree.rf>a#a-ln95046-082-p3s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95046-082-p3s2w14">
     <a>
      <lex.rf>a#a-ln95046-082-p3s2w14</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>uvádět</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>3</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>decl</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w7420f1</val_frame.rf>
     <children>
      <LM id="t-ln95046-082-p3s2w17">
       <a>
        <lex.rf>a#a-ln95046-082-p3s2w17</lex.rf>
        <aux.rf>a#a-ln95046-082-p3s2w16</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>dokument</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln95046-082-p3s2a0">
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Gen</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>2</deepord>
      </LM>
      <LM id="t-ln95046-082-p3s2w4">
       <a>
        <lex.rf>a#a-ln95046-082-p3s2w4</lex.rf>
        <aux.rf>a#a-ln95046-082-p3s2w3</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>mít</t_lemma>
       <functor>EFF</functor>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <quot>
        <LM>
         <type>dsp</type>
         <set_id>ln95046-082-p3s2w4</set_id>
        </LM>
       </quot>
       <is_dsp_root>1</is_dsp_root>
       <sentmod>enunc</sentmod>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>hrt</deontmod>
        <tense>sim</tense>
        <aspect>proc</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w1855f4</val_frame.rf>
       <children>
        <LM id="t-ln95046-082-p3s2w2">
         <a>
          <lex.rf>a#a-ln95046-082-p3s2w2</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>který</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>4</deepord>
         <quot>
          <LM>
           <type>dsp</type>
           <set_id>ln95046-082-p3s2w4</set_id>
          </LM>
         </quot>
         <gram>
          <sempos>n.pron.indef</sempos>
          <gender>anim</gender>
          <number>sg</number>
          <indeftype>total2</indeftype>
          <person>3</person>
         </gram>
        </LM>
        <LM id="t-ln95046-082-p3s2w5">
         <a>
          <lex.rf>a#a-ln95046-082-p3s2w5</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>právo</t_lemma>
         <functor>CPHR</functor>
         <tfa>f</tfa>
         <deepord>7</deepord>
         <quot>
          <LM>
           <type>dsp</type>
           <set_id>ln95046-082-p3s2w4</set_id>
          </LM>
         </quot>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <val_frame.rf>v#v-w4261f1</val_frame.rf>
         <children>
          <LM id="t-ln95046-082-p3s2a2">
           <nodetype>qcomplex</nodetype>
           <is_generated>1</is_generated>
           <t_lemma>#QCor</t_lemma>
           <functor>ACT</functor>
           <tfa>t</tfa>
           <deepord>6</deepord>
           <quot>
            <LM>
             <type>dsp</type>
             <set_id>ln95046-082-p3s2w4</set_id>
            </LM>
           </quot>
           <coref_gram.rf>t-ln95046-082-p3s2w2</coref_gram.rf>
          </LM>
          <LM id="t-ln95046-082-p3s2w6">
           <a>
            <lex.rf>a#a-ln95046-082-p3s2w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>získávat</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>9</deepord>
           <quot>
            <LM>
             <type>dsp</type>
             <set_id>ln95046-082-p3s2w4</set_id>
            </LM>
           </quot>
           <gram>
            <sempos>v</sempos>
            <verbmod>nil</verbmod>
            <deontmod>decl</deontmod>
            <tense>nil</tense>
            <aspect>proc</aspect>
            <resultative>res0</resultative>
            <dispmod>nil</dispmod>
            <iterativeness>it0</iterativeness>
           </gram>
           <val_frame.rf>v#v-w9504f1</val_frame.rf>
           <children>
            <LM id="t-ln95046-082-p3s2a3">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#Cor</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>8</deepord>
             <quot>
              <LM>
               <type>dsp</type>
               <set_id>ln95046-082-p3s2w4</set_id>
              </LM>
             </quot>
             <coref_gram.rf>t-ln95046-082-p3s2w2</coref_gram.rf>
            </LM>
            <LM id="t-ln95046-082-p3s2w7">
             <a>
              <lex.rf>a#a-ln95046-082-p3s2w7</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>obživa</t_lemma>
             <functor>PAT</functor>
             <tfa>f</tfa>
             <deepord>10</deepord>
             <quot>
              <LM>
               <type>dsp</type>
               <set_id>ln95046-082-p3s2w4</set_id>
              </LM>
             </quot>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
            </LM>
            <LM id="t-ln95046-082-p3s2w11">
             <a>
              <lex.rf>a#a-ln95046-082-p3s2w11</lex.rf>
              <aux.rf>a#a-ln95046-082-p3s2w8</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>zaměstnání</t_lemma>
             <functor>LOC</functor>
             <subfunctor>basic</subfunctor>
             <tfa>f</tfa>
             <deepord>11</deepord>
             <quot>
              <LM>
               <type>dsp</type>
               <set_id>ln95046-082-p3s2w4</set_id>
              </LM>
             </quot>
             <gram>
              <sempos>n.denot.neg</sempos>
              <gender>neut</gender>
              <number>sg</number>
              <negation>neg0</negation>
             </gram>
             <children>
              <LM id="t-ln95046-082-p3s2w10">
               <a>
                <lex.rf>a#a-ln95046-082-p3s2w10</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>zvolený</t_lemma>
               <functor>RSTR</functor>
               <tfa>f</tfa>
               <deepord>12</deepord>
               <quot>
                <LM>
                 <type>dsp</type>
                 <set_id>ln95046-082-p3s2w4</set_id>
                </LM>
               </quot>
               <gram>
                <sempos>adj.denot</sempos>
                <degcmp>pos</degcmp>
                <negation>neg0</negation>
               </gram>
               <children>
                <LM id="t-ln95046-082-p3s2w9">
                 <a>
                  <lex.rf>a#a-ln95046-082-p3s2w9</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>svobodný</t_lemma>
                 <functor>MANN</functor>
                 <tfa>f</tfa>
                 <deepord>13</deepord>
                 <quot>
                  <LM>
                   <type>dsp</type>
                   <set_id>ln95046-082-p3s2w4</set_id>
                  </LM>
                 </quot>
                 <gram>
                  <sempos>adj.denot</sempos>
                  <degcmp>pos</degcmp>
                  <negation>neg0</negation>
                 </gram>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln95046-082-p3s3">
   <atree.rf>a#a-ln95046-082-p3s3</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln95046-082-p3s3w7">
     <a>
      <lex.rf>a#a-ln95046-082-p3s3w7</lex.rf>
      <aux.rf>
       <LM>a#a-ln95046-082-p3s3w4</LM>
       <LM>a#a-ln95046-082-p3s3w6</LM>
       <LM>a#a-ln95046-082-p3s3w5</LM>
      </aux.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>zakotvit</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>5</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>cdn</verbmod>
      <deontmod>hrt</deontmod>
      <tense>sim</tense>
      <aspect>cpl</aspect>
      <resultative>res1</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w8885f1</val_frame.rf>
     <children>
      <LM id="t-ln95046-082-p3s3w1">
       <a>
        <lex.rf>a#a-ln95046-082-p3s3w1</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>výkon</t_lemma>
       <functor>PAT</functor>
       <tfa>t</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p3s3w3">
         <a>
          <lex.rf>a#a-ln95046-082-p3s3w3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>politika</t_lemma>
         <functor>APP</functor>
         <tfa>t</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln95046-082-p3s3w2">
           <a>
            <lex.rf>a#a-ln95046-082-p3s3w2</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>sociální</t_lemma>
           <functor>RSTR</functor>
           <tfa>t</tfa>
           <deepord>1</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln95046-082-p3s3a0">
       <nodetype>qcomplex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#Gen</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>4</deepord>
      </LM>
      <LM id="t-ln95046-082-p3s3w9">
       <a>
        <lex.rf>a#a-ln95046-082-p3s3w9</lex.rf>
        <aux.rf>a#a-ln95046-082-p3s3w8</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>sféra</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>6</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln95046-082-p3s3w10">
         <a>
          <lex.rf>a#a-ln95046-082-p3s3w10</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>stát</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94208-140-p1s1">
   <atree.rf>a#a-ln94208-140-p1s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94208-140-p1s1w1">
     <a>
      <lex.rf>a#a-ln94208-140-p1s1w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>hovor</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>inan</gender>
      <number>pl</number>
     </gram>
     <children>
      <LM id="t-ln94208-140-p1s1w3">
       <a>
        <lex.rf>a#a-ln94208-140-p1s1w3</lex.rf>
        <aux.rf>a#a-ln94208-140-p1s1w2</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Lány</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>pl</number>
       </gram>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94208-140-p2s1">
   <atree.rf>a#a-ln94208-140-p2s1</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94208-140-p2s1w1">
     <a>
      <lex.rf>a#a-ln94208-140-p2s1w1</lex.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>výběr</t_lemma>
     <functor>DENOM</functor>
     <tfa>f</tfa>
     <deepord>1</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>n.denot</sempos>
      <gender>inan</gender>
      <number>sg</number>
     </gram>
     <children>
      <LM id="t-ln94208-140-p2s1w4">
       <a>
        <lex.rf>a#a-ln94208-140-p2s1w4</lex.rf>
        <aux.rf>a#a-ln94208-140-p2s1w2</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>rozhovor</t_lemma>
       <functor>DIR1</functor>
       <subfunctor>basic</subfunctor>
       <tfa>f</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94208-140-p2s1w3">
         <a>
          <lex.rf>a#a-ln94208-140-p2s1w3</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>pravidelný</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>3</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
        </LM>
        <LM id="t-ln94208-140-p2s1w8">
         <a>
          <lex.rf>a#a-ln94208-140-p2s1w8</lex.rf>
          <aux.rf>a#a-ln94208-140-p2s1w5</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>Havel</t_lemma>
         <functor>ADDR</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <is_name_of_person>1</is_name_of_person>
         <gram>
          <sempos>n.denot</sempos>
          <gender>anim</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94208-140-p2s1w7">
           <a>
            <lex.rf>a#a-ln94208-140-p2s1w7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>Václav</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>5</deepord>
           <is_name_of_person>1</is_name_of_person>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
          </LM>
          <LM id="t-ln94208-140-p2s1w6">
           <a>
            <lex.rf>a#a-ln94208-140-p2s1w6</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>prezident</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>anim</gender>
            <number>sg</number>
           </gram>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94208-140-p2s1w10">
         <a>
          <lex.rf>a#a-ln94208-140-p2s1w10</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>vysílaný</t_lemma>
         <functor>RSTR</functor>
         <tfa>f</tfa>
         <deepord>7</deepord>
         <gram>
          <sempos>adj.denot</sempos>
          <degcmp>pos</degcmp>
          <negation>neg0</negation>
         </gram>
         <children>
          <LM id="t-ln94208-140-p2s1w12">
           <a>
            <lex.rf>a#a-ln94208-140-p2s1w12</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>rozhlas</t_lemma>
           <functor>ACT</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94208-140-p2s1w11">
             <a>
              <lex.rf>a#a-ln94208-140-p2s1w11</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>Český</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>9</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94208-140-p2s1w15">
         <a>
          <lex.rf>a#a-ln94208-140-p2s1w15</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>být</t_lemma>
         <functor>ID</functor>
         <tfa>f</tfa>
         <deepord>11</deepord>
         <gram>
          <sempos>v</sempos>
          <verbmod>ind</verbmod>
          <deontmod>decl</deontmod>
          <tense>post</tense>
          <aspect>proc</aspect>
          <resultative>res0</resultative>
          <dispmod>disp0</dispmod>
          <iterativeness>it0</iterativeness>
         </gram>
         <val_frame.rf>v#v-w243f2</val_frame.rf>
         <children>
          <LM id="t-ln94208-140-p2s1w14">
           <a>
            <lex.rf>a#a-ln94208-140-p2s1w14</lex.rf>
            <aux.rf>a#a-ln94208-140-p2s1w13</aux.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>#PersPron</t_lemma>
           <functor>LOC</functor>
           <subfunctor>front</subfunctor>
           <tfa>t</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>n.pron.def.pers</sempos>
            <gender>nr</gender>
            <number>pl</number>
            <person>1</person>
            <politeness>basic</politeness>
           </gram>
          </LM>
          <LM id="t-ln94208-140-p2s1w17">
           <a>
            <lex.rf>a#a-ln94208-140-p2s1w17</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>rok</t_lemma>
           <functor>ACT</functor>
           <tfa>f</tfa>
           <deepord>12</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94208-140-p2s1w16">
             <a>
              <lex.rf>a#a-ln94208-140-p2s1w16</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>zajímavý</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>13</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94208-140-p2s2">
   <atree.rf>a#a-ln94208-140-p2s2</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94208-140-p2s2w7">
     <a>
      <lex.rf>a#a-ln94208-140-p2s2w7</lex.rf>
      <aux.rf>
       <LM>a#a-ln94208-140-p2s2w2</LM>
       <LM>a#a-ln94208-140-p2s2w1</LM>
      </aux.rf>
     </a>
     <nodetype>complex</nodetype>
     <t_lemma>účastnit_se</t_lemma>
     <functor>PRED</functor>
     <tfa>f</tfa>
     <deepord>4</deepord>
     <sentmod>enunc</sentmod>
     <gram>
      <sempos>v</sempos>
      <verbmod>ind</verbmod>
      <deontmod>hrt</deontmod>
      <tense>sim</tense>
      <aspect>proc</aspect>
      <resultative>res0</resultative>
      <dispmod>disp0</dispmod>
      <iterativeness>it0</iterativeness>
     </gram>
     <val_frame.rf>v#v-w6995f1</val_frame.rf>
     <children>
      <LM id="t-ln94208-140-p2s2w4">
       <a>
        <lex.rf>a#a-ln94208-140-p2s2w4</lex.rf>
        <aux.rf>a#a-ln94208-140-p2s2w3</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>prosinec</t_lemma>
       <functor>TWHEN</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>1</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln94208-140-p2s2w6">
       <a>
        <lex.rf>a#a-ln94208-140-p2s2w6</lex.rf>
        <aux.rf>a#a-ln94208-140-p2s2w5</aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>Budapešť</t_lemma>
       <functor>LOC</functor>
       <subfunctor>basic</subfunctor>
       <tfa>t</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
      </LM>
      <LM id="t-ln94208-140-p2s2a0">
       <nodetype>complex</nodetype>
       <is_generated>1</is_generated>
       <t_lemma>#PersPron</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>n.pron.def.pers</sempos>
        <gender>nr</gender>
        <number>sg</number>
        <person>1</person>
        <politeness>basic</politeness>
       </gram>
      </LM>
      <LM id="t-ln94208-140-p2s2w8">
       <a>
        <lex.rf>a#a-ln94208-140-p2s2w8</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>summit</t_lemma>
       <functor>PAT</functor>
       <tfa>f</tfa>
       <deepord>5</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>inan</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94208-140-p2s2w9">
         <a>
          <lex.rf>a#a-ln94208-140-p2s2w9</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>konference</t_lemma>
         <functor>APP</functor>
         <tfa>f</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>fem</gender>
          <number>sg</number>
         </gram>
         <val_frame.rf>v#v-w1487f1</val_frame.rf>
         <children>
          <LM id="t-ln94208-140-p2s2w12">
           <a>
            <lex.rf>a#a-ln94208-140-p2s2w12</lex.rf>
           </a>
           <nodetype>coap</nodetype>
           <t_lemma>a</t_lemma>
           <functor>CONJ</functor>
           <deepord>8</deepord>
           <children>
            <LM id="t-ln94208-140-p2s2w11">
             <a>
              <lex.rf>a#a-ln94208-140-p2s2w11</lex.rf>
              <aux.rf>a#a-ln94208-140-p2s2w10</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>bezpečnost</t_lemma>
             <functor>PAT</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>7</deepord>
             <gram>
              <sempos>n.denot.neg</sempos>
              <gender>fem</gender>
              <number>sg</number>
              <negation>neg0</negation>
             </gram>
            </LM>
            <LM id="t-ln94208-140-p2s2w13">
             <a>
              <lex.rf>a#a-ln94208-140-p2s2w13</lex.rf>
              <aux.rf>a#a-ln94208-140-p2s2w10</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>spolupráce</t_lemma>
             <functor>PAT</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>9</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
            </LM>
            <LM id="t-ln94208-140-p2s2w15">
             <a>
              <lex.rf>a#a-ln94208-140-p2s2w15</lex.rf>
              <aux.rf>a#a-ln94208-140-p2s2w14</aux.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>Evropa</t_lemma>
             <functor>LOC</functor>
             <subfunctor>basic</subfunctor>
             <tfa>f</tfa>
             <deepord>10</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94208-140-p2s3">
   <atree.rf>a#a-ln94208-140-p2s3</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94208-140-p2s3w10">
     <a>
      <lex.rf>a#a-ln94208-140-p2s3w10</lex.rf>
     </a>
     <nodetype>coap</nodetype>
     <t_lemma>a</t_lemma>
     <functor>CONJ</functor>
     <deepord>9</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-ln94208-140-p2s3w2">
       <a>
        <lex.rf>a#a-ln94208-140-p2s3w2</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>instituce</t_lemma>
       <functor>ACT</functor>
       <tfa>t</tfa>
       <deepord>2</deepord>
       <gram>
        <sempos>n.denot</sempos>
        <gender>fem</gender>
        <number>sg</number>
       </gram>
       <children>
        <LM id="t-ln94208-140-p2s3w1">
         <a>
          <lex.rf>a#a-ln94208-140-p2s3w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>tento</t_lemma>
         <functor>RSTR</functor>
         <tfa>t</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>adj.pron.def.demon</sempos>
         </gram>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94208-140-p2s3w3">
       <a>
        <lex.rf>a#a-ln94208-140-p2s3w3</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>vzniknout</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>ant</tense>
        <aspect>cpl</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w8671f1</val_frame.rf>
       <children>
        <LM id="t-ln94208-140-p2s3w4">
         <a>
          <lex.rf>a#a-ln94208-140-p2s3w4</lex.rf>
         </a>
         <nodetype>atom</nodetype>
         <t_lemma>ještě</t_lemma>
         <functor>RHEM</functor>
         <tfa>f</tfa>
         <deepord>4</deepord>
        </LM>
        <LM id="t-ln94208-140-p2s3w8">
         <a>
          <lex.rf>a#a-ln94208-140-p2s3w8</lex.rf>
          <aux.rf>
           <LM>a#a-ln94208-140-p2s3w5</LM>
           <LM>a#a-ln94208-140-p2s3w6</LM>
          </aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>rozdělení</t_lemma>
         <functor>TWHEN</functor>
         <subfunctor>basic</subfunctor>
         <tfa>f</tfa>
         <deepord>6</deepord>
         <gram>
          <sempos>n.denot.neg</sempos>
          <gender>neut</gender>
          <number>sg</number>
          <negation>neg0</negation>
         </gram>
         <val_frame.rf>v#v-w5582f2</val_frame.rf>
         <children>
          <LM id="t-ln94208-140-p2s3a0">
           <nodetype>qcomplex</nodetype>
           <is_generated>1</is_generated>
           <t_lemma>#Gen</t_lemma>
           <functor>ACT</functor>
           <tfa>t</tfa>
           <deepord>5</deepord>
          </LM>
          <LM id="t-ln94208-140-p2s3w9">
           <a>
            <lex.rf>a#a-ln94208-140-p2s3w9</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>svět</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>7</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>inan</gender>
            <number>sg</number>
           </gram>
          </LM>
          <LM id="t-ln94208-140-p2s3w7">
           <a>
            <lex.rf>a#a-ln94208-140-p2s3w7</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>bipolární</t_lemma>
           <functor>MANN</functor>
           <tfa>f</tfa>
           <deepord>8</deepord>
           <gram>
            <sempos>adj.denot</sempos>
            <degcmp>pos</degcmp>
            <negation>neg0</negation>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94208-140-p2s3w15">
       <a>
        <lex.rf>a#a-ln94208-140-p2s3w15</lex.rf>
       </a>
       <nodetype>coap</nodetype>
       <t_lemma>#Comma</t_lemma>
       <functor>CONJ</functor>
       <is_member>1</is_member>
       <deepord>14</deepord>
       <children>
        <LM id="t-ln94208-140-p2s3w12">
         <a>
          <lex.rf>a#a-ln94208-140-p2s3w12</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>hledat</t_lemma>
         <functor>PRED</functor>
         <is_member>1</is_member>
         <tfa>f</tfa>
         <deepord>11</deepord>
         <gram>
          <sempos>v</sempos>
          <verbmod>ind</verbmod>
          <deontmod>decl</deontmod>
          <tense>sim</tense>
          <aspect>proc</aspect>
          <resultative>res0</resultative>
          <dispmod>disp0</dispmod>
          <iterativeness>it0</iterativeness>
         </gram>
         <val_frame.rf>v#v-w1058f1</val_frame.rf>
         <children>
          <LM id="t-ln94208-140-p2s3w11">
           <a>
            <lex.rf>a#a-ln94208-140-p2s3w11</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>teď</t_lemma>
           <functor>TWHEN</functor>
           <subfunctor>basic</subfunctor>
           <tfa>t</tfa>
           <deepord>10</deepord>
           <gram>
            <sempos>adv.pron.def</sempos>
           </gram>
          </LM>
          <LM id="t-ln94208-140-p2s3w14">
           <a>
            <lex.rf>a#a-ln94208-140-p2s3w14</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>identita</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>13</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>sg</number>
           </gram>
           <children>
            <LM id="t-ln94208-140-p2s3w13">
             <a>
              <lex.rf>a#a-ln94208-140-p2s3w13</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>#PersPron</t_lemma>
             <functor>APP</functor>
             <tfa>t</tfa>
             <deepord>12</deepord>
             <gram>
              <sempos>n.pron.def.pers</sempos>
              <gender>inher</gender>
              <number>inher</number>
              <person>inher</person>
              <politeness>inher</politeness>
             </gram>
             <coref_gram.rf>t-ln94208-140-p2s3w2</coref_gram.rf>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
        <LM id="t-ln94208-140-p2s3w16">
         <a>
          <lex.rf>a#a-ln94208-140-p2s3w16</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>hledat</t_lemma>
         <functor>PRED</functor>
         <is_member>1</is_member>
         <tfa>t</tfa>
         <deepord>15</deepord>
         <gram>
          <sempos>v</sempos>
          <verbmod>ind</verbmod>
          <deontmod>decl</deontmod>
          <tense>sim</tense>
          <aspect>proc</aspect>
          <resultative>res0</resultative>
          <dispmod>disp0</dispmod>
          <iterativeness>it0</iterativeness>
         </gram>
         <val_frame.rf>v#v-w1058f1</val_frame.rf>
         <children>
          <LM id="t-ln94208-140-p2s3w19">
           <a>
            <lex.rf>a#a-ln94208-140-p2s3w19</lex.rf>
           </a>
           <nodetype>coap</nodetype>
           <t_lemma>#Comma</t_lemma>
           <functor>CONJ</functor>
           <deepord>18</deepord>
           <children>
            <LM id="t-ln94208-140-p2s3w18">
             <a>
              <lex.rf>a#a-ln94208-140-p2s3w18</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>smysl</t_lemma>
             <functor>PAT</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>17</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>inan</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94208-140-p2s3w17">
               <a>
                <lex.rf>a#a-ln94208-140-p2s3w17</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>#PersPron</t_lemma>
               <functor>PAT</functor>
               <tfa>t</tfa>
               <deepord>16</deepord>
               <gram>
                <sempos>n.pron.def.pers</sempos>
                <gender>inher</gender>
                <number>inher</number>
                <person>inher</person>
                <politeness>inher</politeness>
               </gram>
               <coref_gram.rf>t-ln94208-140-p2s3w2</coref_gram.rf>
              </LM>
             </children>
            </LM>
            <LM id="t-ln94208-140-p2s3w20">
             <a>
              <lex.rf>a#a-ln94208-140-p2s3w20</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>vize</t_lemma>
             <functor>PAT</functor>
             <is_member>1</is_member>
             <tfa>f</tfa>
             <deepord>19</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94208-140-p2s3w23">
               <a>
                <lex.rf>a#a-ln94208-140-p2s3w23</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>poslání</t_lemma>
               <functor>PAT</functor>
               <tfa>f</tfa>
               <deepord>21</deepord>
               <gram>
                <sempos>n.denot.neg</sempos>
                <gender>neut</gender>
                <number>sg</number>
                <negation>neg0</negation>
               </gram>
               <val_frame.rf>v#v-w3983f1</val_frame.rf>
               <children>
                <LM id="t-ln94208-140-p2s3w21">
                 <a>
                  <lex.rf>a#a-ln94208-140-p2s3w21</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>#PersPron</t_lemma>
                 <functor>APP</functor>
                 <tfa>t</tfa>
                 <deepord>20</deepord>
                 <gram>
                  <sempos>n.pron.def.pers</sempos>
                  <gender>inher</gender>
                  <number>inher</number>
                  <person>inher</person>
                  <politeness>inher</politeness>
                 </gram>
                 <coref_gram.rf>t-ln94208-140-p2s3w2</coref_gram.rf>
                </LM>
                <LM id="t-ln94208-140-p2s3w22">
                 <a>
                  <lex.rf>a#a-ln94208-140-p2s3w22</lex.rf>
                 </a>
                 <nodetype>complex</nodetype>
                 <t_lemma>budoucí</t_lemma>
                 <functor>RSTR</functor>
                 <tfa>f</tfa>
                 <deepord>22</deepord>
                 <gram>
                  <sempos>adj.denot</sempos>
                  <degcmp>pos</degcmp>
                  <negation>neg0</negation>
                 </gram>
                </LM>
               </children>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
  <LM id="t-ln94208-140-p2s4">
   <atree.rf>a#a-ln94208-140-p2s4</atree.rf>
   <nodetype>root</nodetype>
   <deepord>0</deepord>
   <children>
    <LM id="t-ln94208-140-p2s4w12">
     <a>
      <lex.rf>a#a-ln94208-140-p2s4w12</lex.rf>
     </a>
     <nodetype>coap</nodetype>
     <t_lemma>#Comma</t_lemma>
     <functor>CONJ</functor>
     <deepord>10</deepord>
     <sentmod>enunc</sentmod>
     <children>
      <LM id="t-ln94208-140-p2s4w4">
       <a>
        <lex.rf>a#a-ln94208-140-p2s4w4</lex.rf>
        <aux.rf>
         <LM>a#a-ln94208-140-p2s4w2</LM>
         <LM>a#a-ln94208-140-p2s4w3</LM>
        </aux.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>účastnit_se</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>3</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>hrt</deontmod>
        <tense>sim</tense>
        <aspect>proc</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w6995f1</val_frame.rf>
       <children>
        <LM id="t-ln94208-140-p2s4w1">
         <a>
          <lex.rf>a#a-ln94208-140-p2s4w1</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>potom</t_lemma>
         <functor>TWHEN</functor>
         <subfunctor>basic</subfunctor>
         <tfa>t</tfa>
         <deepord>1</deepord>
         <gram>
          <sempos>adv.pron.def</sempos>
         </gram>
        </LM>
        <LM id="t-ln94208-140-p2s4a1">
         <nodetype>complex</nodetype>
         <is_generated>1</is_generated>
         <t_lemma>#PersPron</t_lemma>
         <functor>ACT</functor>
         <tfa>t</tfa>
         <deepord>2</deepord>
         <gram>
          <sempos>n.pron.def.pers</sempos>
          <gender>nr</gender>
          <number>sg</number>
          <person>1</person>
          <politeness>basic</politeness>
         </gram>
        </LM>
        <LM id="t-ln94208-140-p2s4w6">
         <a>
          <lex.rf>a#a-ln94208-140-p2s4w6</lex.rf>
          <aux.rf>a#a-ln94208-140-p2s4w5</aux.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>leden</t_lemma>
         <functor>TWHEN</functor>
         <subfunctor>basic</subfunctor>
         <tfa>f</tfa>
         <deepord>4</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
        </LM>
        <LM id="t-ln94208-140-p2s4w7">
         <a>
          <lex.rf>a#a-ln94208-140-p2s4w7</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>summit</t_lemma>
         <functor>PAT</functor>
         <tfa>f</tfa>
         <deepord>5</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>inan</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94208-140-p2s4w9">
           <a>
            <lex.rf>a#a-ln94208-140-p2s4w9</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>země</t_lemma>
           <functor>APP</functor>
           <tfa>f</tfa>
           <deepord>6</deepord>
           <gram>
            <sempos>n.denot</sempos>
            <gender>fem</gender>
            <number>pl</number>
           </gram>
           <children>
            <LM id="t-ln94208-140-p2s4w8">
             <a>
              <lex.rf>a#a-ln94208-140-p2s4w8</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>členský</t_lemma>
             <functor>RSTR</functor>
             <tfa>f</tfa>
             <deepord>7</deepord>
             <gram>
              <sempos>adj.denot</sempos>
              <degcmp>pos</degcmp>
              <negation>neg0</negation>
             </gram>
            </LM>
            <LM id="t-ln94208-140-p2s4w10">
             <a>
              <lex.rf>a#a-ln94208-140-p2s4w10</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>Rada</t_lemma>
             <functor>APP</functor>
             <tfa>f</tfa>
             <deepord>8</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>sg</number>
             </gram>
             <children>
              <LM id="t-ln94208-140-p2s4w11">
               <a>
                <lex.rf>a#a-ln94208-140-p2s4w11</lex.rf>
               </a>
               <nodetype>complex</nodetype>
               <t_lemma>bezpečnost</t_lemma>
               <functor>APP</functor>
               <tfa>f</tfa>
               <deepord>9</deepord>
               <gram>
                <sempos>n.denot.neg</sempos>
                <gender>fem</gender>
                <number>sg</number>
                <negation>neg0</negation>
               </gram>
              </LM>
             </children>
            </LM>
           </children>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
      <LM id="t-ln94208-140-p2s4w13">
       <a>
        <lex.rf>a#a-ln94208-140-p2s4w13</lex.rf>
       </a>
       <nodetype>complex</nodetype>
       <t_lemma>být</t_lemma>
       <functor>PRED</functor>
       <is_member>1</is_member>
       <tfa>f</tfa>
       <deepord>11</deepord>
       <gram>
        <sempos>v</sempos>
        <verbmod>ind</verbmod>
        <deontmod>decl</deontmod>
        <tense>post</tense>
        <aspect>proc</aspect>
        <resultative>res0</resultative>
        <dispmod>disp0</dispmod>
        <iterativeness>it0</iterativeness>
       </gram>
       <val_frame.rf>v#v-w243f2</val_frame.rf>
       <children>
        <LM id="t-ln94208-140-p2s4w16">
         <a>
          <lex.rf>a#a-ln94208-140-p2s4w16</lex.rf>
         </a>
         <nodetype>complex</nodetype>
         <t_lemma>výročí</t_lemma>
         <functor>ACT</functor>
         <tfa>f</tfa>
         <deepord>12</deepord>
         <gram>
          <sempos>n.denot</sempos>
          <gender>neut</gender>
          <number>sg</number>
         </gram>
         <children>
          <LM id="t-ln94208-140-p2s4w17">
           <a>
            <lex.rf>a#a-ln94208-140-p2s4w17</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>založení</t_lemma>
           <functor>PAT</functor>
           <tfa>f</tfa>
           <deepord>14</deepord>
           <gram>
            <sempos>n.denot.neg</sempos>
            <gender>neut</gender>
            <number>sg</number>
            <negation>neg0</negation>
           </gram>
           <val_frame.rf>v#v-w8925f1</val_frame.rf>
           <children>
            <LM id="t-ln94208-140-p2s4a0">
             <nodetype>qcomplex</nodetype>
             <is_generated>1</is_generated>
             <t_lemma>#Gen</t_lemma>
             <functor>ACT</functor>
             <tfa>t</tfa>
             <deepord>13</deepord>
            </LM>
            <LM id="t-ln94208-140-p2s4w18">
             <a>
              <lex.rf>a#a-ln94208-140-p2s4w18</lex.rf>
             </a>
             <nodetype>complex</nodetype>
             <t_lemma>OSN</t_lemma>
             <functor>PAT</functor>
             <tfa>f</tfa>
             <deepord>15</deepord>
             <gram>
              <sempos>n.denot</sempos>
              <gender>fem</gender>
              <number>nr</number>
             </gram>
            </LM>
           </children>
          </LM>
          <LM id="t-ln94208-140-p2s4w14">
           <a>
            <lex.rf>a#a-ln94208-140-p2s4w14</lex.rf>
           </a>
           <nodetype>complex</nodetype>
           <t_lemma>50</t_lemma>
           <functor>RSTR</functor>
           <tfa>f</tfa>
           <deepord>16</deepord>
           <gram>
            <sempos>adj.quant.def</sempos>
            <numertype>ord</numertype>
           </gram>
          </LM>
         </children>
        </LM>
       </children>
      </LM>
     </children>
    </LM>
   </children>
  </LM>
 </trees>
</tdata>
